# -*- coding: utf-8 -*-
"""This module contains utilities for seamlessly handling DataFrames as factor matrices and xarray DataArrays as data
"""

__author__ = "Marie Roald & Yngve Mardal Moe"

from functools import wraps
from inspect import signature
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr

from ._module_utils import (
    _SINGLETON,
    _check_is_argument,
    is_dataframe,
    is_xarray,
    validate_cp_tensor,
)

__all__ = [
    "is_dataframe",
    "is_xarray",
    "is_labelled_dataset",
    "is_labelled_cp",
    "is_labelled_tucker",
    "label_cp_tensor",
]


def add_factor_metadata(cp_tensor, dataset):
    """Adds the additional coordinates along each dataset dimension as new columns in the factor matrices.

    The coordinates of xarray DataArrays can contain metadata. For each dimension, there may be additional
    coordinates that are not used for indexing purposes. This function will iterate over all modes of
    a dataset and a labelled CP tensor and add the additional coordinates as new columns in the factor
    matrices.

    Parameters
    ----------
    cp_tensor : labelled CP Tensor
    dataset : xarray.DataArray

    Returns
    -------
    tuple
        CP-tensor like tuple where the factor matrices are augmented with additional metadata.

    Examples
    --------
    >>> from tlviz.data import load_oslo_city_bike
    >>> from tlviz.postprocessing import postprocess, add_factor_metadata
    >>> from tensorly.decomposition import parafac
    >>> bikes = load_oslo_city_bike()
    >>> bikes.coords
    Coordinates:
      * End station name  (End station name) object '7 Juni Plassen' ... 'Økernve...
        lat               (End station name) float64 59.92 59.93 ... 59.93 59.92
        lon               (End station name) float64 10.73 10.75 ... 10.8 10.78
      * Hour              (Hour) int32 0 1 2 3 4 5 6 7 8 ... 16 17 18 19 20 21 22 23
      * Month             (Month) int32 1 2 3 4 5 6 7 8 9 10 11 12
      * Day of week       (Day of week) int32 0 1 2 3 4 5 6
      * Year              (Year) int32 2020 2021

    We see that the ``End station name`` dimension has two additional columns: ``lat`` and ``lon``.
    These contain metadata about the end station coordinates, and it can be useful to have these
    columns also in the factor matrices. To do this, we first fit a PARAFAC model to the dataset,
    then we postprocess it to label the CP tensor and finally, we add the metadata information

    >>> cp = parafac(bikes.data, 3, init="random")
    >>> cp_labelled = postprocess(cp, bikes)
    >>> print(cp_labelled[1][0].columns)
    RangeIndex(start=0, stop=3, step=1)
    >>> cp_with_metadata = add_factor_metadata(cp_labelled, bikes)
    >>> print(cp_with_metadata[1][0].columns)
    Index([0, 1, 2, 'lat', 'lon'], dtype='object')

    We see that when we add the metadata, then the latitude and longitude columns are added
    to the dataframe.
    """
    if not is_labelled_cp(cp_tensor):
        raise ValueError("The CP tensor must be labelled with the same labels as the dataset.")
    if not is_labelled_dataset(dataset):
        raise ValueError("The dataset must be labelled with the same labels as the CP tensor.")

    weights, factor_matrices = cp_tensor
    factors_with_metadata = [None] * len(factor_matrices)
    for mode, factor_matrix in enumerate(factor_matrices):
        dim_name = factor_matrix.index.name
        coords = dataset.coords[dim_name]
        metadata = pd.DataFrame({name: coords.coords[name].to_pandas() for name in coords.coords})
        metadata = metadata.drop(dim_name, axis=1)
        factors_with_metadata[mode] = factor_matrix.join(metadata)
    return weights, factors_with_metadata


def _label_factor_matrices(factor_matrices, dataset):
    if is_xarray(dataset):

        def xarray_to_pandas_index(dataset, dim_name):
            return dataset.coords[dim_name].xindexes[dim_name].to_pandas_index()

        factor_matrices = [
            pd.DataFrame(factor_matrix, index=xarray_to_pandas_index(dataset, dim_name))
            for factor_matrix, dim_name in zip(factor_matrices, dataset.dims)
        ]
    elif is_dataframe(dataset) and len(factor_matrices) == 2:
        factor_matrices = [
            pd.DataFrame(factor_matrices[0], index=dataset.index),
            pd.DataFrame(factor_matrices[1], index=dataset.columns),
        ]
    else:
        raise ValueError(
            "``dataset`` must be xarray.DataArray or, pandas.DataFrame "
            "(only possible if ``len(factor_matrices) == 2``)"
        )
    return factor_matrices


def label_cp_tensor(cp_tensor, dataset):
    """Label the CP tensor by converting the factor matrices into DataFrames with a sensible index.

    Convert the factor matrices into Pandas DataFrames where the DataFrame indices
    are given by the coordinate names of an xarray DataArray. If the dataset has only
    two modes, then it can also be a pandas DataFrame.

    Parameters
    ----------
    cp_tensor : CPTensor
        CP Tensor whose factor matrices should be labelled
    dataset : xarray.DataArray of pandas.DataFrame
        Dataset used to label the factor matrices

    Returns
    -------
    CPTensor
        Tuple on the CPTensor format, except that the factor matrices are DataFrames.
    """
    if is_labelled_cp(cp_tensor) and is_labelled_dataset(dataset):
        warn(
            "Both the CP tensor and the dataset is labelled, the labels from the cp tensor will be overwritten "
            + " with the labels from the dataset."
        )
        cp_tensor = _unlabel_cp_tensor(cp_tensor, optional=False, preserve_columns=True)[0]

    if is_xarray(dataset) or is_dataframe(dataset):
        return (cp_tensor[0], _label_factor_matrices(cp_tensor[1], dataset))
    elif isinstance(dataset, np.ndarray):
        return cp_tensor
    else:
        raise ValueError("Dataset must be either numpy array, xarray or pandas dataframe.")


def get_data(x):
    """Extract the numerical values from ``x`` as a numpy array.

    Arguments
    ---------
    x : np.ndarray or pd.DataFrame or xr.DataArray

    Returns
    -------
    np.ndarray
        The numerical values of ``x`` as a numpy array.
    """
    if is_xarray(x):
        return x.data
    if is_dataframe(x):
        return x.values
    return np.asarray(x)


def is_labelled_cp(cp_tensor):
    """Check if a cp tensor is labelled or not

    Arguments
    ---------
    cp_tensor : tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and an iterable of factor matrices as second argument

    Returns
    -------
    bool
        Whether the factor matrices are labelled or not

    Raises
    ------
    TypeError
        If only some of the factor matrices are labelled (i.e. not none or all).
    """
    num_dataframes = 0
    for factor_matrix in cp_tensor[1]:
        if is_dataframe(factor_matrix):
            num_dataframes += 1

    if num_dataframes == 0:
        return False
    elif num_dataframes == len(cp_tensor[1]):
        return True
    else:
        raise TypeError(
            f"{num_dataframes} out of {len(cp_tensor[1])} factor matrices are labelled (are DataFrames)."
            + " All or none should be labelled."
        )


def is_labelled_tucker(tucker_tensor):
    """Check if a Tucker tensor is labelled or not

    Arguments
    ---------
    tucker_tensor : tuple
        TensorLy-style TuckerTensor object or tuple with the core array as the
        first argument and an iterable of factor matrices as second argument

    Returns
    -------
    bool
        Whether the factor matrices are labelled or not

    Raises
    ------
    TypeError
        If only some of the factor matrices are labelled (i.e. not none or all).
    """
    return is_labelled_cp(tucker_tensor)  # The weights are not considered for cp, neither is the core array for tucker


def is_labelled_dataset(x):
    """Returns True if the dataset is labelled (is a DataFrame or DataArray).

    This function is the same as writing ``is_dataframe(x) or is_xarray(x)``.

    Parameters
    ----------
    x
        Variable to check

    Returns
    -------
    bool
        Whether ``x`` is labelled or not.

    """
    # TOTEST: is_labeled_dataset
    return is_dataframe(x) or is_xarray(x)


def _extract_df_metadata(df, preserve_columns=True):
    values = df.values
    if preserve_columns:
        metadata = {"index": df.index, "columns": df.columns}
    else:
        metadata = {"index": df.index}
    return values, metadata


# TODO: Make public?
def _unlabel_cp_tensor(cp_tensor, optional, preserve_columns):
    if cp_tensor is None and optional:
        return None, None
    elif cp_tensor is None:
        raise TypeError("cp_tensor cannot be None")
    weights, factors = cp_tensor

    # Check that factor matrices are valid
    is_labelled = is_dataframe(factors[0])
    for factor in factors:
        if is_dataframe(factor) != is_labelled:
            raise ValueError("All factor matrices must either be labelled or not labelled.")

    if not is_labelled:
        return (weights, factors), None

    unlabelled_factors = []
    factor_metadata = []
    for factor in factors:
        factor, metadata = _extract_df_metadata(factor, preserve_columns=preserve_columns)
        unlabelled_factors.append(factor)
        factor_metadata.append(metadata)
    return (weights, unlabelled_factors), factor_metadata


def _relabel_cp_tensor(cp_tensor, factor_metadata, optional):
    if cp_tensor is None and optional:
        return

    if factor_metadata is None:
        return cp_tensor

    weights, factors = cp_tensor
    labelled_factors = []
    for factor, metadata in zip(factors, factor_metadata):
        labelled_factors.append(pd.DataFrame(factor, **metadata))
    return weights, labelled_factors


def _unlabel_factor_matrix(factor_matrix, optional, preserve_columns):
    if factor_matrix is None and optional:
        return None, None
    if not is_dataframe(factor_matrix):
        return factor_matrix, None
    return _extract_df_metadata(factor_matrix, preserve_columns=preserve_columns)


def _relabel_factor_matrix(factor_matrix, factor_metadata, optional):
    if factor_matrix is None and optional:
        return

    if factor_metadata is None:
        return factor_matrix

    return pd.DataFrame(factor_matrix, **factor_metadata)


def _unlabel_dataset(dataset, optional):
    if optional and dataset is None:
        return None, None, None
    elif dataset is None:  # Not optional and dataset is None
        raise TypeError("Dataset cannot be None")
    if is_xarray(dataset):
        np_dataset = dataset.values
        dataset_constructor = xr.DataArray
        dataset_metadata = {
            "name": dataset.name,
            "coords": dataset.coords,
            "dims": dataset.dims,
            "attrs": dataset.attrs,
        }
    elif is_dataframe(dataset):
        np_dataset = dataset.values
        dataset_constructor = pd.DataFrame
        dataset_metadata = {
            "index": dataset.index,
            "columns": dataset.columns,
        }
    else:
        np_dataset = dataset
        dataset_constructor = np.array
        dataset_metadata = {}
    return np_dataset, dataset_constructor, dataset_metadata


def _relabel_dataset(np_dataset, dataset_constructor, dataset_metadata, optional):
    if optional and np_dataset is None:
        return
    return dataset_constructor(np_dataset, **dataset_metadata)


def _handle_labelled_cp(cp_tensor_name, output_cp_tensor_index, optional=False, preserve_columns=True):
    def decorator(func):
        _check_is_argument(func, cp_tensor_name)

        @wraps(func)
        def func2(*args, **kwargs):
            bound_arguments = signature(func).bind(*args, **kwargs)

            cp_tensor = bound_arguments.arguments.get(cp_tensor_name, None)
            if cp_tensor is not None:
                validate_cp_tensor(cp_tensor)

            cp_tensor_unlabelled, cp_tensor_metadata = _unlabel_cp_tensor(
                cp_tensor, optional=optional, preserve_columns=preserve_columns
            )

            bound_arguments.arguments[cp_tensor_name] = cp_tensor_unlabelled
            out = func(*bound_arguments.args, **bound_arguments.kwargs)

            if output_cp_tensor_index is _SINGLETON:
                out = _relabel_cp_tensor(out, cp_tensor_metadata, optional=optional)
            elif output_cp_tensor_index is not None:
                out_cp_tensor = _relabel_cp_tensor(out[output_cp_tensor_index], cp_tensor_metadata, optional=optional)
                out = (
                    *out[:output_cp_tensor_index],
                    out_cp_tensor,
                    *out[output_cp_tensor_index + 1 :],
                )
            return out

        return func2

    return decorator


def _handle_labelled_dataset(dataset_name, output_dataset_index, optional=False):
    def decorator(func):
        _check_is_argument(func, dataset_name)

        @wraps(func)
        def func2(*args, **kwargs):
            bound_arguments = signature(func).bind(*args, **kwargs)

            if optional and dataset_name not in bound_arguments.arguments:
                return func(*bound_arguments.args, **bound_arguments.kwargs)
            dataset = bound_arguments.arguments[dataset_name]
            dataset_unlabelled, dataset_constructor, dataset_metadata = _unlabel_dataset(dataset, optional=optional)

            bound_arguments.arguments[dataset_name] = dataset_unlabelled
            out = func(*bound_arguments.args, **bound_arguments.kwargs)

            if output_dataset_index is _SINGLETON:
                out = _relabel_dataset(out, dataset_constructor, dataset_metadata, optional=optional)
            elif output_dataset_index is not None:
                out_dataset = _relabel_dataset(
                    out[output_dataset_index], dataset_constructor, dataset_metadata, optional=optional
                )
                out = (
                    *out[:output_dataset_index],
                    out_dataset,
                    *out[output_dataset_index + 1 :],
                )
            return out

        return func2

    return decorator


def _handle_labelled_factor_matrix(
    factor_matrix_name, output_factor_matrix_index, optional=False, preserve_columns=True
):
    def decorator(func):
        _check_is_argument(func, factor_matrix_name)

        @wraps(func)
        def func2(*args, **kwargs):
            bound_arguments = signature(func).bind(*args, **kwargs)

            factor_matrix = bound_arguments.arguments.get(factor_matrix_name, None)

            factor_matrix_unlabelled, factor_matrix_metadata = _unlabel_factor_matrix(
                factor_matrix, optional=optional, preserve_columns=preserve_columns
            )

            bound_arguments.arguments[factor_matrix_name] = factor_matrix_unlabelled
            out = func(*bound_arguments.args, **bound_arguments.kwargs)

            if output_factor_matrix_index is _SINGLETON:
                out = _relabel_factor_matrix(out, factor_matrix_metadata, optional=optional)
            elif output_factor_matrix_index is not None:
                out_factor_matrix = _relabel_factor_matrix(
                    out[output_factor_matrix_index], factor_matrix_metadata, optional=optional,
                )
                out = (
                    *out[:output_factor_matrix_index],
                    out_factor_matrix,
                    *out[output_factor_matrix_index + 1 :],
                )
            return out

        return func2

    return decorator



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

from warnings import warn

import numpy as np
import scipy.linalg as sla

from . import factor_tools
from ._module_utils import is_iterable
from ._tl_utils import _handle_tensorly_backends_cp, _handle_tensorly_backends_dataset
from ._xarray_wrapper import (
    _SINGLETON,
    _handle_labelled_cp,
    _handle_labelled_dataset,
    add_factor_metadata,
    label_cp_tensor,
)
from .utils import unfold_tensor

__all__ = [
    "resolve_cp_sign_indeterminacy",
    "postprocess",
    "factor_matrix_to_tidy",
    "add_factor_metadata",
    "label_cp_tensor",
]


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", _SINGLETON)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", _SINGLETON)
def resolve_cp_sign_indeterminacy(cp_tensor, dataset, resolve_mode=None, unresolved_mode=-1, method="transpose"):
    r"""Resolve the sign indeterminacy of CP models.

    Tensor factorisations have a sign indeterminacy that allows any change in the
    sign of the component vectors in one mode, under the condition that the sign
    of a component vector in another mode changes as well. This means that we can
    "flip" any component vector so long as the corresponding component vector in
    another mode is also flipped. This flipping can hurt the model's interpretability.
    For example, if a factor represents a chemical spectrum, then this flipping may lead
    to it being negative instead of positive.

    To illustrate the sign indeterminacy, we start with the SVD, which is on the form

    .. math::

        \mathbf{X} = \mathbf{U} \mathbf{S} \mathbf{V}^\mathsf{T}.

    The factorisation above is equivalent with the following factorisation:

    .. math::
        \mathbf{X} = (\mathbf{U} \text{diag}(\mathbf{f})) \mathbf{S} (\mathbf{V} \text{diag}(\mathbf{f}))^\mathsf{T},

    where :math:`\mathbf{f}` is a vector containing only ones or negative ones. Similarly,
    a CP factorisation with factor matrices :math:`\mathbf{A}, \mathbf{B}` and :math:`\mathbf{C}`
    is equivalent to the CP factorisations with the following factor matrices:

     * :math:`(\mathbf{A} \text{diag}(\mathbf{f})), (\mathbf{B} \text{diag}(\mathbf{f}))` and :math:`\mathbf{C}`
     * :math:`(\mathbf{A} \text{diag}(\mathbf{f})), \mathbf{B}` and :math:`(\mathbf{C} \text{diag}(\mathbf{f}))`
     * :math:`\mathbf{A}, (\mathbf{B} \text{diag}(\mathbf{f}))` and :math:`(\mathbf{C} \text{diag}(\mathbf{f}))`

    One way to circumvent the sign indeterminacy is by imposing non-negativity. However,
    that is not always a reasonable choice (e.g. if the data also contains negative entries).
    When we don't want to impose non-negativity constraints, then we need some other way to
    resolve the sign indeterminacy (which this function provides). The idea is easiest described
    in the two-way (matrix) case.

    Consider a data matrix, :math:`\mathbf{X}` whose columns represent samples and rows represent
    measurements. Then, we want the measurement-mode component-vectors to be mostly aligned with
    the data matrix. The components should describe what the data is, not what it is not.
    For example, if the data is non-negative, then the measurement-mode component vectors should
    be mostly non-negative. With the SVD, we can compute whether we should flip the :math:`r`-th
    column of :math:`\mathbf{U}` by computing

    .. math::

        f_r = \sum_{i=1^I} v_{ir}^2 \text{sign}{v_{ir}}

    if :math:`f_r` is negative, then we should flip the sign of the :math:`r-th` column of
    :math:`\mathbf{U}` and :math:`\mathbf{V}` :cite:p:`bro2008resolving`.

    The methodology above works well in practice, and is rooted in the fact that the
    :math:`i`-th row of :math:`\mathbf{V}` can be interpreted as the coordinates of the
    :math:`i`-th row of :math:`\mathbf{X}` in a vector space spanned by the columns of
    :math:`\mathbf{U}`. Then, the above equation will give us component vectors where the data
    points is mainly located in the non-negative orthant.

    The above interpretation is correct under the assumption: :math:`\mathbf{U}^\mathsf{T}\mathbf{U} = \mathbf{I}`.
    However, the heuristic still works well when this is not the case :cite:p:`bro2013solving`.
    Still, we also include a modification of the above scheme where the same interpretation
    holds with non-orthogonal factors.

    .. math::

        f_r = \sum_{i=1}^I h_{ir}^2 \text{sign}(h_{ir}),

    where :math:`\mathbf{H} = \mathbf{U}(\mathbf{U}^\mathsf{T}\mathbf{U})^{-1} \mathbf{X}`.
    That is the rows of :math:`\mathbf{H}` represent the rows of :math:`\mathbf{X}` as
    described by the column basis of :math:`\mathbf{U}`.

    In the multiway case, when :math:`\mathcal{X}` is a tensor instead of a matrix, we can
    apply the same logic :cite:p:`bro2013solving`. If we have the factor matrices
    :math:`\mathbf{A}, \mathbf{B}` and :math:`\mathbf{C}`, then we flip the sign of any
    factor matrix (e.g. :math:`\mathbf{A}`) by computing

    .. math::

        f_r^{(\mathbf{A})} = \sum_{i=1}^I {h_{ir}^{(\mathbf{A})}}^2 \text{sign}({h_{ir}^{(\mathbf{A})}}),

    where :math:`\mathbf{H}^{(\mathbf{A})} = \mathbf{A}^\mathsf{T} \mathbf{X}_{(0)}` or
    :math:`\mathbf{H}^{(\mathbf{A})} = \mathbf{A}(\mathbf{A}^\mathsf{T}\mathbf{A})^{-1} \mathbf{X}_{(0)}`,
    depending on whether the scheme based on the SVD scheme :cite:p:`bro2013solving` or the
    corrected scheme. :math:`\mathbf{X}_{(0)} \in \mathbb{R}^{I \times JK}` is the tensor,
    :math:`\mathcal{X}`, unfolded along the first mode. We can then correct the sign of
    :math:`\mathbf{A}` by multiplying and one of the other factor matrices by
    :math:`\text{diag}(\mathbf{f}^{(\mathbf{A})})`. By using this procedure, we can align all
    factor matrices except for one (the unresolved mode) with the "direction of the data".

    Note that this sign indeterminacy comes as a direct consequence of the scaling indeterminacy
    of component models, since :math:`\text{diag}(\mathbf{f})^{-1} = \text{diag}(\mathbf{f})`.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.
    resolve_mode : int, iterable or None
        Mode(s) whose factor matrix should be aligned with the data. If
        None, then the sign should be corrected for all modes except the
        ``unresolved_mode``.
    unresolved_mode : int
        Mode used to correct the sign indeterminacy in other mode(s). The
        factor matrix in this mode may not be aligned with the data.
    method : "transpose" or "positive_coord"
        Which method to use when computing the signs. Use ``"transpose"``
        for the method in :cite:p:`bro2008resolving,bro2013solving`, and
        ``"positive_coord"`` for the method corrected for non-orthogonal
        factor matrices described above.

    Returns
    -------
    CPTensor or tuple
        The CP tensor after correcting the signs.

    Raises
    ------
    ValueError
        If ``unresolved_mode`` is not between ``-dataset.ndim`` and ``dataset.ndim-1``.
    ValueError
        If ``unresolved_mode`` is in ``resolve_mode``

    Notes
    -----
    For more information, see :cite:p:`bro2008resolving,bro2013solving`
    """
    if unresolved_mode < 0:
        unresolved_mode = dataset.ndim + unresolved_mode
    if unresolved_mode >= dataset.ndim or unresolved_mode < 0:
        raise ValueError("`unresolved_mode` must be between `-dataset.ndim` and `dataset.ndim-1`.")
    if resolve_mode == unresolved_mode or (is_iterable(resolve_mode) and unresolved_mode in resolve_mode):
        raise ValueError("The unresolved mode cannot be resolved.")

    if resolve_mode is None:
        resolve_mode = range(dataset.ndim)

    if is_iterable(resolve_mode):
        for mode in resolve_mode:
            if mode != unresolved_mode:
                cp_tensor = resolve_cp_sign_indeterminacy(
                    cp_tensor, dataset, unresolved_mode=unresolved_mode, resolve_mode=mode, method=method,
                )

        return cp_tensor

    unfolded_dataset = unfold_tensor(dataset, resolve_mode)
    factor_matrix = cp_tensor[1][resolve_mode]

    if method == "transpose":
        sign_scores = factor_matrix.T @ unfolded_dataset
    elif method == "positive_coord":
        sign_scores = sla.lstsq(factor_matrix, unfolded_dataset)[0]
    else:
        raise ValueError("Method must be either `transpose` or `positive_coord`")

    signs = np.sign(np.sum(sign_scores ** 2 * np.sign(sign_scores), axis=1))
    signs = np.asarray(signs).reshape(1, -1)

    factor_matrices = list(cp_tensor[1])
    factor_matrices[resolve_mode] = factor_matrices[resolve_mode] * signs
    factor_matrices[unresolved_mode] = factor_matrices[unresolved_mode] * signs
    return cp_tensor[0], tuple(factor_matrices)


@_handle_tensorly_backends_cp("reference_cp_tensor", None, optional=True)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_cp("reference_cp_tensor", None, optional=True)
def postprocess(
    cp_tensor,
    dataset=None,
    reference_cp_tensor=None,
    permute=True,
    resolve_mode=None,
    unresolved_mode=-1,
    flip_method="transpose",
    weight_behaviour="normalise",
    weight_mode=0,
    allow_smaller_rank=False,
    include_metadata=False,
):
    """Standard postprocessing of a CP decomposition.

    This function will perform standard postprocessing of a CP decomposition.
    If a reference CP tensor is provided, then the columns of ``cp_tensor``'s
    factor matrices are aligned with the columns of ``reference_cp_tensor``'s
    factor matrices.

    Next, the factor matrices of the CP tensor are scaled according the the specified
    weight behaviour (default is normalised).

    If a dataset is provided, then the sign indeterminacy is resolved and if the
    dataset is labelled (i.e. is an xarray or a dataframe), then the factor matrices
    of the CP tensor is labelled too.

    This function is equivalent to calling

     * ``permute_cp_tensor``
     * ``distribute_weights``
     * ``resolve_cp_sign_indeterminacy``
     * ``label_cp_tensor``

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        CPTensor to postprocess
    dataset : ndarray or xarray (optional)
        Dataset the CP tensor represents
    reference_cp_tensor : CPTensor or tuple (optional)
        If provided, then the tensor whose factors we align the CP tensor's
        columns with.
    permute : bool
        If ``True``, then the factors are permuted in descending weight order if
        ``reference_cp_tensor`` is ``None``.
    resolve_mode : int, iterable or None
        Mode(s) whose factor matrix should be aligned with the data. If
        None, then the sign should be corrected for all modes except the
        ``unresolved_mode``.
    unresolved_mode : int
        Mode used to correct the sign indeterminacy in other mode(s). The
        factor matrix in this mode may not be aligned with the data.
    method : "transpose" or "positive_coord"
        Which method to use when computing the signs. Use ``"transpose"``
        for the method in :cite:p:`bro2008resolving,bro2013solving`, and
        ``"positive_coord"`` for the method corrected for non-orthogonal
        factor matrices described above.
    weight_behaviour : {"ignore", "normalise", "evenly", "one_mode"} (default="normalise")
        How to handle the component weights.

         * ``"ignore"`` - Do nothing
         * ``"normalise"`` - Normalise all factor matrices
         * ``"evenly"`` - All factor matrices have equal norm
         * ``"one_mode"`` - The weight is allocated in one mode, all other factor matrices have unit norm columns.

    weight_mode : int (optional)
        Which mode to have the component weights in (only used if ``weight_behaviour="one_mode"``)
    allow_smaller_rank : bool (default=False)
        If ``True``, then a low rank decomposition can be permuted against one with higher rank. The "missing columns"
        are padded by nan values
    include_metadata : bool (default=True)
        If ``True``, then the factor metadata will be added as columns in the factor matrices.


    Returns
    -------
    CPTensor
        The post processed CPTensor.

    See Also
    --------
    tlviz.factor_tools.permute_cp_tensor
    tlviz.factor_tools.distribute_weights
    tlviz.postprocessing.resolve_cp_sign_indeterminacy
    tlviz.postprocessing.label_cp_tensor

    Examples
    --------

    Here is an example were we use postprocess on a decomposition of aminoacid data

    .. plot::
        :context: close-figs
        :include-source:

        >>> import tlviz
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>> dataset = tlviz.data.load_aminoacids()
        Loading Aminoacids dataset from:
        Bro, R, PARAFAC: Tutorial and applications, Chemometrics and Intelligent Laboratory Systems, 1997, 38, 149-171

        The dataset is an xarray DataArray and it contains relevant side information

        >>> print(type(dataset))
        <class 'xarray.core.dataarray.DataArray'>

        We see that after postprocessing, the cp_tensor contains pandas DataFrames

        >>> cp_tensor = parafac(dataset.data, 3, init="random", random_state=0)
        >>> cp_tensor_postprocessed = tlviz.postprocessing.postprocess(cp_tensor, dataset)
        >>> print(type(cp_tensor[1][0]))
        <class 'numpy.ndarray'>
        >>> print(type(cp_tensor_postprocessed[1][0]))
        <class 'pandas.core.frame.DataFrame'>


        We see that after postprocessing, the factor matrix has unit norm

        >>> print(np.linalg.norm(cp_tensor[1][0], axis=0))
        [160.82985402 182.37338941 125.3689186 ]
        >>> print(np.linalg.norm(cp_tensor_postprocessed[1][0], axis=0))
        [1. 1. 1.]

        When we construct a dense tensor from a postprocessed cp_tensor it is constructed
        as an xarray DataArray

        >>> print(type(tlviz.utils.cp_to_tensor(cp_tensor)))
        <class 'numpy.ndarray'>
        >>> print(type(tlviz.utils.cp_to_tensor(cp_tensor_postprocessed)))
        <class 'xarray.core.dataarray.DataArray'>

        The visualisation of the postprocessed cp_tensor shows that the scaling and sign indeterminacy
        is taken care of and x-xaxis has correct labels and ticks

        >>> fig, ax = tlviz.visualisation.components_plot(cp_tensor)
        >>> plt.show()

        >>> fig, ax = tlviz.visualisation.components_plot(cp_tensor_postprocessed)
        >>> plt.show()

    """
    if not permute and reference_cp_tensor is not None:
        warn("``permute=False`` is ignored if a reference CP tensor is provided.")

    if permute or reference_cp_tensor is not None:
        cp_tensor = factor_tools.permute_cp_tensor(
            cp_tensor, reference_cp_tensor=reference_cp_tensor, allow_smaller_rank=allow_smaller_rank
        )

    cp_tensor = factor_tools.distribute_weights(cp_tensor, weight_behaviour=weight_behaviour, weight_mode=weight_mode)

    if dataset is not None:
        cp_tensor = resolve_cp_sign_indeterminacy(
            cp_tensor, dataset, resolve_mode=resolve_mode, unresolved_mode=unresolved_mode, method=flip_method,
        )

        cp_tensor = label_cp_tensor(cp_tensor, dataset)

    if include_metadata and dataset is not None:
        cp_tensor = add_factor_metadata(cp_tensor, dataset)
    elif include_metadata:
        warn("Cannot include metadata when there is no provided dataset")

    return cp_tensor


def factor_matrix_to_tidy(factor_matrix, var_name="Component", value_name="Signal", id_vars=None):
    """Convert a factor matrix into a tidy dataset, for use with Plotly Express.

    If we convert a factor matrix into a tidy dataset (or long table), then we get a table on the form

    .. list-table:: Tidy format factor matrix
        :widths: 25 25 25
        :header-rows: 1

        * - Index
          - Component
          - Signal
        * - 0
          - 0
          - 0.1
        * - 1
          - 0
          - 0.5
        * - ...
          - ...
          - ...
        * - 38
          - 2
          - 0.7
        * - 39
          - 2
          - 0.2

    The component vectors are all stacked on top of each other, with a separate column that specifies which
    component each row belongs to. This function can also preserve metadata, which is signified by columns
    that have non-integer column names. For example, if we have a dataframe on the form

    .. list-table:: Factor matrix with metadata
        :widths: 25 25 25 25 25 25
        :header-rows: 1

        * - Index
          - 0
          - 1
          - 2
          - lat
          - lon
        * - 0
          - 0.1
          - 0.2
          - 0.5
          - 59.91273
          - 10.74609
        * - 1
          - 0.5
          - 0.2
          - 0.1
          - 63.43049
          - 10.39506
        * - ...
          - ...
          - ...
          - ...
          - ...
          - ...
        * - 5
          - 0.2
          - 0.1
          - 0.3
          - 60.39299
          - 5.32415
        * - 5
          - 0.0
          - 0.2
          - 0.1
          - 58.97005
          - 5.73332

    and convert it into a tidy format factor matrix, then we get a table on the form

    .. list-table:: Tidy format factor matrix with metadata
        :widths: 25 25 25 25 25
        :header-rows: 1

        * - Index
          - lat
          - lon
          - Component
          - Signal
        * - 0
          - 59.91273
          - 10.74609
          - 0
          - 0.1
        * - 1
          - 63.43049
          - 10.39506
          - 0
          - 0.5
        * - ...
          - ...
          - ...
          - ...
          - ...
        * - 4
          - 69.6489
          - 18.95508
          - 2
          - 0.0
        * - 5
          - 58.97005
          - 5.73332
          - 2
          - 0.1

    Parameters
    ----------
    factor_matrix : pd.DataFrame
        A labelled factor matrix potentially with metadata columns
    var_name : str
        Name of the column that specifies which component each row belongs to
    value_name : str
        Name of the column that holds the magnitude of each component entry
    id_vars : iterable or None (default=None)
        Which columns to interpret as metadata. The columns not specified here are considered as the components.
        If ``id_vars is None``, then all columns with non-integer names are considered metadata columns.

    Returns
    -------
    pd.DataFrame
        Tidy format factor matrix

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from tlviz.postprocessing import factor_matrix_to_tidy
    >>> rng = np.random.default_rng(0)
    >>> factor_matrix = pd.DataFrame(rng.uniform(size=(10, 3)))
    >>> factor_matrix.head()
              0         1         2
    0  0.636962  0.269787  0.040974
    1  0.016528  0.813270  0.912756
    2  0.606636  0.729497  0.543625
    3  0.935072  0.815854  0.002739
    4  0.857404  0.033586  0.729655
    >>> tidy_factor_matrix = factor_matrix_to_tidy(factor_matrix)
    >>> tidy_factor_matrix.head()
       index Component    Signal
    0      0         0  0.636962
    1      1         0  0.016528
    2      2         0  0.606636
    3      3         0  0.935072
    4      4         0  0.857404
    >>> factor_matrix_with_metadata = factor_matrix.copy()
    >>> factor_matrix_with_metadata["Metadata"] = rng.uniform(size=10)
    >>> factor_matrix_with_metadata.head()
              0         1         2  Metadata
    0  0.636962  0.269787  0.040974  0.688447
    1  0.016528  0.813270  0.912756  0.388921
    2  0.606636  0.729497  0.543625  0.135097
    3  0.935072  0.815854  0.002739  0.721488
    4  0.857404  0.033586  0.729655  0.525354
    >>> tidy_factor_matrix_with_metadata = factor_matrix_to_tidy(factor_matrix_with_metadata)
    >>> tidy_factor_matrix_with_metadata.head()
       Metadata  index Component    Signal
    0  0.688447      0         0  0.636962
    1  0.388921      1         0  0.016528
    2  0.135097      2         0  0.606636
    3  0.721488      3         0  0.935072
    4  0.525354      4         0  0.857404
    """
    factor_matrix = factor_matrix.reset_index()
    if id_vars is None:
        id_vars = set()
        for column in factor_matrix.columns:
            if type(column) != int:
                id_vars.add(column)
    id_vars = sorted(id_vars)

    return factor_matrix.melt(var_name=var_name, value_name=value_name, id_vars=id_vars)



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"
__version__ = "0.1.3"


from . import (
    data,
    factor_tools,
    model_evaluation,
    multimodel_evaluation,
    outliers,
    postprocessing,
    utils,
    visualisation,
)



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

from warnings import warn

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm
from matplotlib.lines import Line2D

from . import factor_tools, model_evaluation, postprocessing
from ._module_utils import _handle_none_weights_cp_tensor, is_dataframe, is_iterable
from ._tl_utils import _handle_tensorly_backends_cp, _handle_tensorly_backends_dataset, to_numpy_cp
from ._xarray_wrapper import _handle_labelled_cp, _handle_labelled_dataset
from .model_evaluation import estimate_core_tensor
from .outliers import (
    _LEVERAGE_NAME,
    _SLABWISE_SSE_NAME,
    compute_outlier_info,
    get_leverage_outlier_threshold,
    get_slabwise_sse_outlier_threshold,
)
from .utils import _alias_mode_axis, cp_to_tensor

__all__ = [
    "scree_plot",
    "histogram_of_residuals",
    "residual_qq",
    "outlier_plot",
    "component_scatterplot",
    "core_element_plot",
    "core_element_heatmap",
    "components_plot",
    "component_comparison_plot",
    "optimisation_diagnostic_plots",
    "percentage_variation_plot",
]


def scree_plot(cp_tensors, dataset, errors=None, metric="Fit", ax=None):
    """Create scree plot for the given cp tensors.

    A scree plot is a plot with the model on the x-axis and a metric (often fit)
    on the y-axis. It is commonly plotted as a line plot with a scatter point located at each
    model.

    Parameters
    ----------
    cp_tensor: dict[Any, CPTensor]
        Dictionary mapping model names (often just the number of components as an int) to a
        model.
    dataset: numpy.ndarray or xarray.DataArray
        Dataset to compare the model against.
    errors: dict[Any, float] (optional)
        The metric to plot. If given, then the cp_tensor and dataset-arguments are ignored.
        This is useful to save computation time if, for example, the fit is computed beforehand.
    metric: str or Callable
        Which metric to plot, should have the signature ``metric(cp_tensor, dataset)`` and return
        a float. If it is a string, then this will be used as the y-label and metric will be set to
        ``metric = getattr(tlviz.model_evaluation, metric)``.
        Also, if ``metric`` is a string, then it is converted to lower-case letters and spaces
        are converted to underlines before getting the metric from the ``model_evaluation`` module.
    ax: matplotlib axes
        Matplotlib axes that the plot will be placed in. If ``None``, then ``plt.gca()`` will be used.

    Returns
    -------
    ax
        Matplotlib axes object with the scree plot


    Examples
    --------
    Simple scree plot of fit

    .. plot ::
        :context: close-figs
        :include-source:

        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import scree_plot
        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>>
        >>> dataset = simulated_random_cp_tensor((10, 20, 30), rank=3, noise_level=0.2, seed=42)[1]
        >>> cp_tensors = {}
        >>> for rank in range(1, 5):
        ...     cp_tensors[f"{rank} components"] = parafac(dataset, rank, random_state=1)
        >>>
        >>> ax = scree_plot(cp_tensors, dataset)
        >>> plt.show()

    Scree plots for fit and core consistency in the same figure

    .. plot ::
        :context: close-figs
        :include-source:

        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import scree_plot
        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>>
        >>> dataset = simulated_random_cp_tensor((10, 20, 30), rank=3, noise_level=0.2, seed=42)[1]
        >>> cp_tensors = {}
        >>> for rank in range(1, 5):
        ...     cp_tensors[rank] = parafac(dataset, rank, random_state=1)
        >>>
        >>> fig, axes = plt.subplots(1, 2, figsize=(8, 2), tight_layout=True)
        >>> ax = scree_plot(cp_tensors, dataset, ax=axes[0])
        >>> ax = scree_plot(cp_tensors, dataset, metric="Core consistency", ax=axes[1])
        >>> # Names are converted to lowercase and spaces are converted to underlines when fetching metric-function,
        >>> # so "Core consistency" becomes getattr(tlviz.model_evaluation, "core_consistency")
        >>>
        >>> for ax in axes:
        ...     xlabel = ax.set_xlabel("Number of components")
        ...     xticks = ax.set_xticks(list(cp_tensors.keys()))
        >>> limits = axes[1].set_ylim((0, 105))
        >>> plt.show()
    """
    if ax is None:
        ax = plt.gca()

    if isinstance(metric, str):
        ax.set_ylabel(metric.replace("_", " "))
        metric = getattr(model_evaluation, metric.lower().replace(" ", "_"))
    cp_tensors = dict(cp_tensors)

    if errors is None:
        # compute error using the metric function
        errors = {model: metric(cp_tensor, dataset) for model, cp_tensor in cp_tensors.items()}
    else:
        errors = dict(errors)

    ax.plot(errors.keys(), errors.values(), "-o")
    return ax


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def histogram_of_residuals(cp_tensor, dataset, ax=None, standardised=True, **kwargs):
    r"""Create a histogram of model residuals (:math:`\hat{\mathbf{\mathcal{X}}} - \mathbf{\mathcal{X}}`).

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        Dataset to compare with
    ax : Matplotlib axes (Optional)
        Axes to plot the histogram in
    standardised : bool
        If true, then the residuals are divided by their standard deviation
    **kwargs
        Additional keyword arguments passed to the histogram function

    Returns
    -------
    ax : Matplotlib axes

    Examples
    --------

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import histogram_of_residuals
        >>> true_cp, X = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
        >>> est_cp = parafac(X, 3)
        >>> histogram_of_residuals(est_cp, X)
        <AxesSubplot:title={'center':'Histogram of residuals'}, xlabel='Standardised residuals', ylabel='Frequency'>
        >>> plt.show()
    """
    estimated_dataset = cp_to_tensor(cp_tensor)
    residuals = (estimated_dataset - dataset).ravel()

    if ax is None:
        ax = plt.gca()
    if standardised:
        residuals = residuals / np.std(residuals)
        ax.set_xlabel("Standardised residuals")
    else:
        ax.set_xlabel("Residuals")

    ax.hist(residuals, **kwargs)
    ax.set_ylabel("Frequency")
    ax.set_title("Histogram of residuals")

    return ax


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def residual_qq(cp_tensor, dataset, ax=None, use_pingouin=False, **kwargs):
    """QQ-plot of the model residuals.

    By default, ``statsmodels`` is used to create the QQ-plot. However,
    if ``use_pingouin=True``, then we import the GPL-3 lisenced Pingouin
    library to create a more informative QQ-plot.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        Dataset to compare with
    ax : Matplotlib axes (Optional)
        Axes to plot the qq-plot in
    use_pingouin : bool
        If true, then the GPL-3 licensed ``pingouin``-library will be used
        for generating an enhanced QQ-plot (with error bars), at the cost
        of changing the license of tlviz into a GPL-license too.
    **kwargs
        Additional keyword arguments passed to the qq-plot function
        (``statsmodels.api.qqplot`` or ``pingouin.qqplot``)

    Returns
    -------
    ax : Matplotlib axes

    Examples
    --------

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import residual_qq
        >>> true_cp, X = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
        >>> est_cp = parafac(X, 3)
        >>> residual_qq(est_cp, X)
        <AxesSubplot:title={'center':'QQ-plot of residuals'}, xlabel='Theoretical Quantiles', ylabel='Sample Quantiles'>
        >>> plt.show()
    """
    estimated_dataset = cp_to_tensor(cp_tensor)
    residuals = (estimated_dataset - dataset).ravel()

    if ax is None:
        ax = plt.gca()

    if use_pingouin:  # pragma: no cover
        from pingouin import qqplot

        warn("GPL-3 Lisenced code is loaded, so this code also follows the GPL-3 license.")
        qqplot(residuals, ax=ax, **kwargs)
    else:
        sm.qqplot(residuals, ax=ax, **kwargs)

    ax.set_title("QQ-plot of residuals")
    return ax


@_handle_tensorly_backends_cp("cp_tensor", None)
@_alias_mode_axis()
def outlier_plot(
    cp_tensor,
    dataset,
    mode=0,
    leverage_rules_of_thumb=None,
    residual_rules_of_thumb=None,
    p_value=0.05,
    ax=None,
    axis=None,  # Alias for mode
):
    """Create the leverage-residual scatterplot to detect outliers.

    Detecting outliers can be a difficult task, and a common way to do this is by making a scatter-plot where the
    leverage score is plotted against the slabwise SSE (or residual).

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        Dataset to compare with
    mode : int
        Which mode (axis) to create the outlier plot for
    leverage_rules_of_thumb : str or iterable of str
        Rule of thumb(s) used to create lines for detecting outliers based on leverage score. Must be a supported
        argument for ``method`` with :meth:`tlviz.outliers.get_leverage_outlier_threshold`. If
        ``leverage_rules_of_thumb`` is an iterable of strings, then multiple lines will be drawn, one for each
        method.
    residual_rules_of_thumb : str or iterable of str
        Rule of thumb(s) used to create lines for detecting outliers based on residuals. Must be a supported
        argument for ``method`` with :meth:`tlviz.outliers.get_slabwise_sse_outlier_threshold`. If
        ``residual_rules_of_thumb`` is an iterable of strings, then multiple lines will be drawn, one for each
        method.
    p_value : float or iterable of float
        p-value(s) to use for both the leverage and residual rules of thumb. If an iterable of floats is used,
        then there will be drawn lines for each p-value.
    ax : Matplotlib axes
        Axes to plot outlier plot in. If ``None``, then ``plt.gca()`` is used.
    axis : int (optional)
        Alias for mode. If set, then mode cannot be set.

    Returns
    -------
    ax : Matplotlib axes
        Axes with outlier plot in

    Examples
    --------
    Here is a simple example demonstrating how to use the outlier plot to detect outliers based on the
    Oslo bike sharing data.

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import non_negative_parafac_hals
        >>> from tlviz.data import load_oslo_city_bike
        >>> from tlviz.postprocessing import postprocess
        >>> from tlviz.visualisation import outlier_plot
        >>>
        >>> data = load_oslo_city_bike()
        >>> X = data.data
        >>> cp = non_negative_parafac_hals(X, 3, init="random")
        >>> cp = postprocess(cp, dataset=data, )
        >>>
        >>> outlier_plot(
        ...     cp, data, leverage_rules_of_thumb='p-value', residual_rules_of_thumb='p-value', p_value=[0.05, 0.01]
        ... )
        <AxesSubplot:title={'center':'Outlier plot for End station name'}, xlabel='Leverage score', ylabel='Slabwise SSE'>
        >>> plt.show()

    We can also provide multiple types of rules of thumb

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import non_negative_parafac_hals
        >>> from tlviz.data import load_oslo_city_bike
        >>> from tlviz.postprocessing import postprocess
        >>> from tlviz.visualisation import outlier_plot
        >>>
        >>> data = load_oslo_city_bike()
        >>> X = data.data
        >>> cp = non_negative_parafac_hals(X, 3, init="random")
        >>> cp = postprocess(cp, dataset=data, )
        >>>
        >>> outlier_plot(
        ...     cp, data, leverage_rules_of_thumb=['huber lower', 'hw higher'], residual_rules_of_thumb='two sigma'
        ... )
        <AxesSubplot:title={'center':'Outlier plot for End station name'}, xlabel='Leverage score', ylabel='Slabwise SSE'>
        >>> plt.show()

    See Also
    --------
    tlviz.outliers.compute_outlier_info
    tlviz.outliers.compute_leverage
    tlviz.outliers.compute_slabwise_sse
    tlviz.outliers.get_leverage_outlier_threshold
    tlviz.outliers.get_slabwise_sse_outlier_threshold
    """
    weights, factor_matrices = cp_tensor

    outlier_info = compute_outlier_info(cp_tensor, dataset, mode=mode)

    if ax is None:
        ax = plt.gca()

    ax.plot(outlier_info[f"{_LEVERAGE_NAME}"], outlier_info[f"{_SLABWISE_SSE_NAME}"], "o", zorder=1, alpha=0.8)
    ax.set_xlabel("Leverage score")
    ax.set_ylabel("Slabwise SSE")
    if is_dataframe(factor_matrices[mode]) and factor_matrices[mode].index.name not in {None, ""}:
        title = f"Outlier plot for {factor_matrices[mode].index.name}"
    else:
        title = f"Outlier plot for mode {mode}"
    ax.set_title(title)

    for x, y, s in zip(outlier_info[f"{_LEVERAGE_NAME}"], outlier_info[f"{_SLABWISE_SSE_NAME}"], outlier_info.index,):
        ax.text(x, y, s, zorder=0)
    # Vertical lines for leverage based rule-of-thumb thresholds
    leverage_thresholds = {}
    if leverage_rules_of_thumb is not None:
        if isinstance(leverage_rules_of_thumb, str):
            leverage_rules_of_thumb = [leverage_rules_of_thumb]

        for leverage_rule_of_thumb in leverage_rules_of_thumb:
            if leverage_rule_of_thumb in {"p-value", "hotelling"} and not is_iterable(p_value):
                p_values = [p_value]
            elif leverage_rule_of_thumb in {"p-value", "hotelling"}:
                p_values = p_value
            else:
                p_values = [None]  # We still need something to iterate over even if it doesn't use the p-value

            for p in p_values:
                threshold = get_leverage_outlier_threshold(
                    outlier_info[f"{_LEVERAGE_NAME}"], method=leverage_rule_of_thumb, p_value=p,
                )

                if leverage_rule_of_thumb == "p-value":
                    name = f"Leverage p-value: {p}"
                elif leverage_rule_of_thumb == "hotelling":
                    name = f"Hotelling T2 p-value: {p}"
                else:
                    name = leverage_rule_of_thumb
                leverage_thresholds[name] = threshold

    # Draw the lines
    for key, value in leverage_thresholds.items():
        ax.axvline(value, label=key, **next(ax._get_lines.prop_cycler))

    residual_thresholds = {}
    if residual_rules_of_thumb is not None:
        if isinstance(residual_rules_of_thumb, str):
            residual_rules_of_thumb = [residual_rules_of_thumb]

        for residual_rule_of_thumb in residual_rules_of_thumb:
            if residual_rule_of_thumb == "p-value" and not is_iterable(p_value):
                p_values = [p_value]
            elif residual_rule_of_thumb == "p-value":
                p_values = p_value
            else:
                p_values = [None]  # We still need something to iterate over even if it doesn't use the p-value

            for p in p_values:
                threshold = get_slabwise_sse_outlier_threshold(
                    outlier_info[f"{_SLABWISE_SSE_NAME}"], method=residual_rule_of_thumb, p_value=p
                )
                if residual_rule_of_thumb == "p-value":
                    name = f"Residual p-value: {p}"
                else:
                    name = residual_rule_of_thumb
                residual_thresholds[name] = threshold
    for key, value in residual_thresholds.items():
        ax.axhline(value, label=key, **next(ax._get_lines.prop_cycler))

    if len(leverage_thresholds) > 0 or len(residual_thresholds) > 0:
        ax.legend()
    return ax


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_none_weights_cp_tensor("cp_tensor")
@_alias_mode_axis()
def component_scatterplot(cp_tensor, mode, x_component=0, y_component=1, ax=None, axis=None, **kwargs):
    """Scatterplot of two columns in a factor matrix.

    Create a scatterplot with the columns of a factor matrix as feature-vectors.
    Note that since factor matrices are not orthogonal, the distances between points
    can be misleading. The lack of orthogonality means that distances and angles
    are "skewed", and two slabs with vastly different locations in the scatter plot
    can be very similar (in the case of collinear components). For more information
    about this phenomenon, see :cite:p:`kiers2000some` and example 8.3 in
    :cite:p:`smilde2005multi`.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    mode : int
        Mode for the factor matrix whose columns are plotted
    x_component : int
        Component plotted on the x-axis
    y_component : int
        Component plotted on the y-axis
    ax : Matplotlib axes (Optional)
        Axes to plot the scatterplot in
    axis : int (optional)
        Alias for mode. If this is provided, then no value for mode can be provided.
    **kwargs
        Additional keyword arguments passed to ``ax.scatter``.

    Returns
    -------
    ax : Matplotlib axes

    Examples
    --------
    Small example with a simulated third order CP tensor

    .. plot::
        :context: close-figs
        :include-source:

        >>> from tensorly.random import random_cp
        >>> from tlviz.visualisation import component_scatterplot
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor = random_cp(shape=(5,10,15), rank=2)
        >>> component_scatterplot(cp_tensor, mode=0)
        <AxesSubplot:title={'center':'Component plot'}, xlabel='Component 0', ylabel='Component 1'>
        >>> plt.show()

    Eexample with PCA of a real stock dataset

    .. plot::
        :context: close-figs
        :include-source:

        >>> import pandas as pd
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import plotly.express as px
        >>> from tlviz.postprocessing import label_cp_tensor
        >>> from tlviz.visualisation import component_scatterplot
        >>>
        >>> # Load data and convert to xarray
        >>> stocks = px.data.stocks().set_index("date").stack()
        >>> stocks.index.names = ["Date", "Stock"]
        >>> stocks = stocks.to_xarray()
        >>>
        >>> # Compute PCA via SVD of centered data
        >>> stocks -= stocks.mean(axis=0)
        >>> U, s, Vh = np.linalg.svd(stocks, full_matrices=False)
        >>>
        >>> # Extract two components and convert to cp_tensor
        >>> num_components = 2
        >>> cp_tensor = s[:num_components], (U[:, :num_components], Vh.T[:, :num_components])
        >>> cp_tensor = label_cp_tensor(cp_tensor, stocks)
        >>>
        >>> # Visualise the components with components_plot
        >>> component_scatterplot(cp_tensor, mode=1)
        <AxesSubplot:title={'center':'Component plot'}, xlabel='Component 0', ylabel='Component 1'>
        >>> plt.show()
    """
    if ax is None:
        ax = plt.gca()

    factor_matrix = cp_tensor[1][mode]
    if is_dataframe(factor_matrix):
        index = factor_matrix.index
        factor_matrix = factor_matrix.values
    else:
        index = np.arange(factor_matrix.shape[0])

    relevant_factors = factor_matrix[:, [x_component, y_component]]

    ax.set_xlabel(f"Component {x_component}")
    ax.set_ylabel(f"Component {y_component}")
    ax.set_title("Component plot")
    ax.scatter(relevant_factors[:, 0], relevant_factors[:, 1], **kwargs)

    for x, y, s in zip(relevant_factors[:, 0], relevant_factors[:, 1], index):
        ax.text(x, y, s)

    return ax


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_none_weights_cp_tensor("cp_tensor")
def core_element_plot(cp_tensor, dataset, normalised=False, ax=None):
    """Scatter plot with the elements of the optimal core tensor for a given CP tensor.

    If the CP-model is appropriate for the data, then the core tensor should
    be superdiagonal, and all off-superdiagonal entries should be zero. This plot
    shows the core elements, sorted so the first R scatter-points correspond to the
    superdiagonal and the subsequent scatter-points correspond to off-diagonal entries
    in the optimal core tensor.

    Together with the scatter plot, there is a line-plot that indicate where the scatter-points
    should be if the CP-model perfectly describes the data.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        The dataset the CP tensor models.
    normalised : bool
        If true then the normalised core consistency will be estimated
        (see ``tlviz.model_evaluation.core_consistency``)
    ax : Matplotlib axes
        Axes to plot the core element plot within

    Returns
    -------
    ax : Matplotlib axes

    Examples
    --------

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import core_element_plot
        >>> true_cp, X = simulated_random_cp_tensor((10, 20, 30), 3, seed=42)
        >>> est_cp = parafac(X, 3)
        >>> core_element_plot(est_cp, X)
        <AxesSubplot:title={'center':'Core consistency: 99.8'}, xlabel='Core element', ylabel='Value'>
        >>> plt.show()
    """
    weights, factors = cp_tensor
    rank = weights.shape[0]

    A = factors[0].copy()
    if weights is not None:
        A *= weights
    factors = tuple((A, *factors[1:]))

    # Estimate core and compute core consistency
    core_tensor = estimate_core_tensor(factors, dataset)
    T = np.zeros([rank] * dataset.ndim)
    np.fill_diagonal(T, 1)
    if normalised:
        denom = np.sum((core_tensor) ** 2)
    else:
        denom = rank

    core_consistency = 100 - 100 * np.sum((core_tensor - T) ** 2) / denom

    # Extract superdiagonal and offdiagonal elements
    core_elements = np.zeros_like(core_tensor.ravel())
    diagonal_mask = np.zeros([rank] * dataset.ndim, dtype=np.bool)
    np.fill_diagonal(diagonal_mask, 1)

    core_elements[:rank] = core_tensor[diagonal_mask]
    core_elements[rank:] = core_tensor[~diagonal_mask]

    # Plot core elements
    if ax is None:
        ax = plt.gca()

    x = np.arange(len(core_elements))
    y = np.zeros_like(x)
    y[:rank] = 1
    ax.plot(x, y, "-", label="Target")
    ax.plot(x[:rank], core_elements[:rank], "o", label="Superdiagonal")
    ax.plot(x[rank:], core_elements[rank:], "x", label="Off diagonal")
    ax.legend()
    ax.set_xlabel("Core element")
    ax.set_ylabel("Value")
    ax.set_title(f"Core consistency: {core_consistency:.1f}")

    return ax


def _srgb_to_luminance(srgb):
    """Return the Y of XYZ.

    Computed based on a preview of the IEC 61966-2-1:1999/AMD1:2003 standard. Downloaded
    from https://www.sis.se/api/document/preview/562720/ (archived version:
    https://web.archive.org/web/20200608215908/https://www.sis.se/api/document/preview/562720/).

    Arguments
    ---------
    srgb : np.ndarray
        Non-linear sRGB signal

    Returns
    -------
    np.ndarray
        Array with luminance (Y) values.
    """
    srgb_linear = np.where(srgb < 0.04045, srgb / 12.92, ((srgb + 0.055) / 1.055) ** 2.4)
    return srgb_linear @ np.array([0.2126, 0.7152, 0.0722])


def _get_core_tensor_index(slab_idx, slice_mode):
    slices = []
    slice_strs = []
    for mode in range(3):
        if mode == slice_mode:
            slices.append(slab_idx)
            slice_strs.append(str(slab_idx))
        else:
            slices.append(slice(None))
            slice_strs.append(":")
    slice_str = ", ".join(slice_strs)

    return tuple(slices), slice_str


def _apply_cmap(selected_slab, vmax):
    cmap = cm.get_cmap("coolwarm")
    scaled_slab = (selected_slab + vmax) / (2 * vmax)
    scaled_slab[scaled_slab > 1] = 1
    scaled_slab[scaled_slab < 0] = 0
    return cmap(scaled_slab)[..., :-1]


def _get_text_color(bg_rgb):
    luminance = _srgb_to_luminance(bg_rgb)
    if luminance > 0.408:
        return "0.15"
    else:
        return "white"


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_none_weights_cp_tensor("cp_tensor")
def core_element_heatmap(cp_tensor, dataset, slice_mode=0, vmax=None, annotate=True, text_kwargs=None, text_fmt=".2f"):
    """Create a heatmap of the slabs of the optimal core tensor for a given CP tensor and dataset.

    It can be useful look at the optimal core tensor for a given CP tensor. This can give valuable information about which
    components that are modelling multi-linear behaviour and which are not. For example, a component that models noise is
    more likely to have strong interactions with the other components compared to a component that have a meaningful
    interpretation. In the core element heatmap, this is shown as rows, columns and/or slabs that have high entries compared
    to the diagonal.

    If the data follows a PARAFAC model perfectly, then there should only be one non-zero entry per slice. For the :math:`r`-th
    slice, the :math:`(r, r)`-th entry will be 1 and all others will be 0.

    .. note::

        The core element heatmap can only be plotted for third-order tensors.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        The dataset the CP tensor models.
    slice_mode : {0, 1, 2} (default=0)
        Which mode to slice the core tensor across.
    vmax : float (default=None)
        The maximum value for the colormap (a diverging colormap with center at 0 will be used).
        If ``None``, then the maximum entry in the core tensor is used.
    annotate : bool (default=True)
        If ``True``, then the value of the core tensor is plotted too.
    text_kwargs : dict (default=None)
        Additional keyword arguments used for plotting the text. Can for example be used to set the font size.
    text_fmt : str (default=".2f")
        Formatting string used for annotating.

    Returns
    -------
    fig : matplotlib.figure.Figure
    axes : ndarray(dtype=matplotlib.axes.Axes)


    Examples
    --------
    .. plot::
        :context: close-figs
        :include-source:

        >>> from tlviz.visualisation import core_element_heatmap
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor, dataset = simulated_random_cp_tensor((20, 30, 40), 3, seed=0)
        >>> fig, axes = core_element_heatmap(cp_tensor, dataset)
        >>> plt.show()
    """
    weights, factors = cp_tensor
    if len(factors) != 3:
        raise ValueError("Can only create a core element heatmap for third order tensors.")

    # Multiply weights into components so diagonal should be one
    A = factors[0].copy()
    A *= weights
    factors = tuple((A, *factors[1:]))

    # Estimate core tensor
    core_tensor = estimate_core_tensor(factors, dataset)
    num_components = core_tensor.shape[0]

    fig, axes = plt.subplots(1, num_components, figsize=(3 * num_components, 3), sharex=True, sharey=True)

    if vmax is None:
        vmax = np.abs(core_tensor).max()

    if text_kwargs is None:
        text_kwargs = {}

    for slab, ax in enumerate(axes):
        slices, slice_str = _get_core_tensor_index(slab, slice_mode)
        selected_slab = core_tensor[slices]
        image = _apply_cmap(selected_slab, vmax)
        ax.imshow(image)

        if annotate:
            for index, value in np.ndenumerate(selected_slab):
                ax.text(
                    *index[::-1],  # Reverse since matrix index is (y, x), not (x, y)
                    f"{value:{text_fmt}}",
                    verticalalignment="center",
                    horizontalalignment="center",
                    color=_get_text_color(image[index]),
                    **text_kwargs,
                )

        ax.set_xticks(np.arange(num_components))
        ax.set_yticks(np.arange(num_components))
        ax.set_title(f"core_tensor[{slice_str}]")

    return fig, axes


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_none_weights_cp_tensor("cp_tensor")
def components_plot(cp_tensor, weight_behaviour="normalise", weight_mode=0, plot_kwargs=None):
    """Plot the component vectors of a CP model.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    weight_behaviour : {'ignore', 'normalise', 'evenly', 'one_mode'}
        How to handle the component weights.

         * ignore - Do nothing, just plot the factor matrices
         * normalise - Plot all components after normalising them
         * evenly - Distribute the weight evenly across all modes
         * one_mode - Move all the weight into one factor matrix

    weight_mode : int
        The mode that the weight should be placed in (only used if ``weight_behaviour='one_mode'``)
    plot_kwargs : list of dictionaries
        List of same length as the number of modes. Each element is a kwargs-dict passed to
        the plot function for that mode.

    Returns
    -------
    fig : matplotlib.figure.Figure
    axes : ndarray(dtype=matplotlib.axes.Axes)

    Examples
    --------
    Small example with a simulated CP tensor

    .. plot::
        :context: close-figs
        :include-source:

        >>> from tensorly.random import random_cp
        >>> from tlviz.visualisation import components_plot
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor = random_cp(shape=(5,10,15), rank=3)
        >>> fig, axes = components_plot(cp_tensor)
        >>> plt.show()

    Full example with PCA of a real stock dataset

    .. plot::
        :context: close-figs
        :include-source:

        >>> import pandas as pd
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import plotly.express as px
        >>> from tlviz.postprocessing import label_cp_tensor
        >>> from tlviz.visualisation import components_plot
        >>>
        >>> # Load data and convert to xarray
        >>> stocks = px.data.stocks().set_index("date").stack()
        >>> stocks.index.names = ["Date", "Stock"]
        >>> stocks = stocks.to_xarray()
        >>>
        >>> # Compute PCA via SVD of centered data
        >>> stocks -= stocks.mean(axis=0)
        >>> U, s, Vh = np.linalg.svd(stocks, full_matrices=False)
        >>>
        >>> # Extract two components and convert to cp_tensor
        >>> num_components = 2
        >>> cp_tensor = s[:num_components], (U[:, :num_components], Vh.T[:, :num_components])
        >>> cp_tensor = label_cp_tensor(cp_tensor, stocks)
        >>>
        >>> # Visualise the components with components_plot
        >>> fig, axes = components_plot(cp_tensor, weight_behaviour="one_mode", weight_mode=1,
        ...                             plot_kwargs=[{}, {'marker': 'o', 'linewidth': 0}])
        >>> plt.show()
    """
    weights, factor_matrices = factor_tools.distribute_weights(
        cp_tensor, weight_behaviour=weight_behaviour, weight_mode=weight_mode
    )

    num_components = len(weights.reshape(-1))
    num_modes = len(factor_matrices)

    if plot_kwargs is None:
        plot_kwargs = [{}] * num_modes

    fig, axes = plt.subplots(1, num_modes, figsize=(16, 9 / num_modes), tight_layout=True)

    for mode, factor_matrix in enumerate(factor_matrices):
        if hasattr(factor_matrix, "plot"):
            factor_matrix.plot(ax=axes[mode], **plot_kwargs[mode])
        else:
            axes[mode].plot(factor_matrix, **plot_kwargs[mode])
            axes[mode].set_xlabel(f"Mode {mode}")
            axes[mode].legend([str(i) for i in range(num_components)])
    return fig, axes


def component_comparison_plot(
    cp_tensors, row="model", weight_behaviour="normalise", weight_mode=0, plot_kwargs=None,
):
    """Create a plot to compare different CP tensors.

    This function creates a figure with either D columns and R rows or D columns and N rows,
    where D is the number of modes, R is the number of components and N is the number of cp tensors
    to compare.

    Parameters
    ----------
    cp_tensors : dict (str -> CPTensor)
        Dictionary with model names mapping to decompositions. The model names
        are used for labels. The components of all CP tensors will be aligned
        to maximise the factor match score with the components of the first CP
        tensor in the dictionary (starting with Python 3.7, dictionaries are sorted by
        insertion order).
    row : {"model", "component"}
    weight_behaviour : {"ignore", "normalise", "evenly", "one_mode"} (default="normalise")
        How to handle the component weights.

         * ``"ignore"`` - Do nothing
         * ``"normalise"`` - Normalise all factor matrices
         * ``"evenly"`` - All factor matrices have equal norm
         * ``"one_mode"`` - The weight is allocated in one mode, all other factor matrices have unit norm columns.

    weight_mode : int (optional)
        Which mode to have the component weights in (only used if ``weight_behaviour="one_mode"``)
    plot_kwargs : list of list of dicts
        Nested list of dictionaries, one dictionary with keyword arguments for each subplot.

    Returns
    -------
    fig : matplotlib figure
    axes : array of matplotlib axes

    Examples
    --------

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tensorly.decomposition import parafac, non_negative_parafac_hals
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.visualisation import component_comparison_plot
        >>> from tlviz.postprocessing import postprocess
        >>>
        >>> true_cp, X = simulated_random_cp_tensor((10, 20, 30), 3, noise_level=0.5, seed=42)
        >>> cp_tensors = {
        ...     "True": true_cp,
        ...     "CP": parafac(X, 3),
        ...     "NN CP": non_negative_parafac_hals(X, 3),
        ... }
        >>> fig, axes = component_comparison_plot(cp_tensors, row="component")
        >>> plt.show()

    If not all decompositions have the same number of components, then the components will be aligned
    with the first (reference) decomposition in the ``cp_tensors``-dictionary. If one of the subsequent
    decompositions have fewer components than the reference decomposition, then the columns will be
    aligned correctly, and if one of them has more, then the additional components will be ignored.

    .. plot::
        :context: close-figs
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> from tlviz.factor_tools import permute_cp_tensor
        >>> from tlviz.postprocessing import postprocess
        >>> from tlviz.visualisation import component_comparison_plot
        >>>
        >>> four_components = simulated_random_cp_tensor((5, 6, 7), 4, noise_level=0.5, seed=42)[0]
        >>> three_components = permute_cp_tensor(four_components, permutation=[0, 1, 2])
        >>> two_components = permute_cp_tensor(four_components, permutation=[0, 2])
        >>> # Plot the decomposition
        >>> cp_tensors = {
        ...     "True": three_components,  # Reference decomposition
        ...     "subset": two_components,  # Only component 0 and 2
        ...     "superset": four_components,  # All components in reference plus one additional
        ... }
        >>> fig, axes = component_comparison_plot(cp_tensors, row="model")
        >>> plt.show()
    """
    cp_tensors = {key: to_numpy_cp(cp_tensor) for key, cp_tensor in cp_tensors.items()}
    main_cp_tensor = next(iter(cp_tensors.values()))
    weights, factor_matrices = main_cp_tensor

    cp_tensors = {
        key: postprocessing.postprocess(
            value,
            reference_cp_tensor=main_cp_tensor,
            weight_behaviour=weight_behaviour,
            weight_mode=weight_mode,
            allow_smaller_rank=True,
        )
        for key, value in cp_tensors.items()
    }

    num_components = factor_matrices[0].shape[1]
    num_modes = len(factor_matrices)
    num_models = len(cp_tensors)
    ref_name = next(iter(cp_tensors.keys()))

    if row == "model":
        num_rows = num_models
    elif row == "component":
        num_rows = num_components
    else:
        raise ValueError("Row must be either 'model' or 'component'")

    fig, axes = plt.subplots(num_rows, num_modes, figsize=(16, num_rows * 9 / num_modes), tight_layout=True)
    for model_num, (model_name, cp_tensor) in enumerate(cp_tensors.items()):
        factor_matrices = cp_tensor[1]  # The weights are handled by the above postprocessing
        if factor_matrices[0].shape[1] > num_components:
            warn(
                f"The {model_name} decomposition has a higher rank than the reference {ref_name} decomposition."
                + f" Therefore, only the subset of columns in {model_name} that correspond to columns in"
                + f" {ref_name} will be plotted."
            )
        for mode, factor_matrix in enumerate(factor_matrices):
            for component_num in range(num_components):
                if row == "model":
                    row_idx = model_num
                elif row == "component":
                    row_idx = component_num

                if plot_kwargs is None:
                    kwargs = {}
                else:
                    kwargs = plot_kwargs[row_idx][mode]

                if is_dataframe(factor_matrix):
                    factor_matrix[component_num].plot(ax=axes[row_idx, mode], **kwargs)
                    legend = axes[row_idx, mode].get_legend()
                    if legend is not None:
                        legend.remove()
                else:
                    axes[row_idx, mode].plot(factor_matrix[:, component_num], **kwargs)
                    axes[row_idx, mode].set_xlabel(f"Mode {mode}")

    if row == "model":
        fig.legend(
            [f"Component {i}" for i in range(num_components)], loc="upper center", ncol=num_components,
        )
        for row_idx, model_name in enumerate(cp_tensors):
            axes[row_idx, 0].set_ylabel(model_name)
    elif row == "component":
        fig.legend(cp_tensors.keys(), loc="upper center", ncol=len(cp_tensors))
        for row_idx in range(num_components):
            axes[row_idx, 0].set_ylabel(f"Component {row_idx}")

    for row_idx in range(num_rows - 1):
        for mode in range(num_modes):
            ax = axes[row_idx, mode]
            xlim = ax.get_xlim()  # Necessary to supress FixedLocator warning
            ax.set_xticks(ax.get_xticks())  # Necessary to supress FixedLocator warning
            ax.set_xticklabels(["" for _ in ax.get_xticks()])
            ax.set_xlim(xlim)  # Necessary to supress FixedLocator warning
            ax.set_xlabel("")
    return fig, axes


def optimisation_diagnostic_plots(error_logs, n_iter_max):
    """Diagnostic plots for the optimisation problem.

    This function creates two plots. One plot that shows the loss value for each initialisation
    and whether or not that initialisation converged or ran until the maximum number of iterations.
    The other plot shows the error log for each initialisation, with the initialisation with lowest
    final error in a different colour (orange).

    These plots can be helpful for understanding how stable the model is with respect to initialisation.
    Ideally, we should see that many initialisations converged and obtained the same, low, error.
    If models converge, but with different errors, then this can indicate that indicates that a stricter
    convergence tolerance is required, and if no models converge, then more iterations may be required.

    Parameters
    ----------
    error_logs : list of arrays
        List of arrays, each containing the error per iteration for an initialisation.
    n_iter_max : int
        Maximum number of iterations for the fitting procedure. Used to determine if the
        models converged or not.

    Returns
    -------
    fig : matplotlib.figure.Figure
    axes : array(dtype=matplotlib.axes.Axes)

    Examples
    --------
    Fit the wrong number of components to show local minima problems

    .. plot::
        :context: close-figs
        :include-source:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from tensorly.random import random_cp
        >>> from tensorly.decomposition import parafac
        >>> from tlviz.visualisation import optimisation_diagnostic_plots
        >>>
        >>> # Generate random tensor and add noise
        >>> rng = np.random.RandomState(1)
        >>> cp_tensor = random_cp((5, 6, 7), 2, random_state=rng)
        >>> dataset = cp_tensor.to_tensor() + rng.standard_normal((5, 6, 7))
        >>>
        >>> # Fit 10 models
        >>> errs = []
        >>> for i in range(10):
        ...     errs.append(parafac(dataset, 3, n_iter_max=500, return_errors=True, init="random", random_state=rng)[1])
        >>>
        >>> # Plot the diganostic plots
        >>> fig, axes = optimisation_diagnostic_plots(errs, 500)
        >>> plt.show()


    Fit a model with too few iterations

    .. plot::
        :context: close-figs
        :include-source:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from tensorly.random import random_cp
        >>> from tensorly.decomposition import parafac
        >>> from tlviz.visualisation import optimisation_diagnostic_plots
        >>>
        >>> # Generate random tensor and add noise
        >>> rng = np.random.RandomState(1)
        >>> cp_tensor = random_cp((5, 6, 7), 3, random_state=rng)
        >>> dataset = cp_tensor.to_tensor() + rng.standard_normal((5, 6, 7))
        >>>
        >>> # Fit 10 models
        >>> errs = []
        >>> for i in range(10):
        ...     errs.append(parafac(dataset, 3, n_iter_max=50, return_errors=True, init="random", random_state=rng)[1])
        >>>
        >>> # Plot the diagnostic plots
        >>> fig, axes = optimisation_diagnostic_plots(errs, 50)
        >>> plt.show()
    """
    fig, axes = plt.subplots(1, 2, figsize=(16, 4.5), tight_layout=True)

    selected_init = None
    lowest_error = np.inf
    for init, error in enumerate(error_logs):
        if error[-1] < lowest_error:
            selected_init = init
            lowest_error = error[-1]

    ymax = 0
    for init, error in enumerate(error_logs):
        if init == selected_init:
            alpha = 1
            color = plt.rcParams["axes.prop_cycle"].by_key()["color"][1]
            zorder = 10
        else:
            alpha = 0.5
            color = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]
            zorder = 0

        if len(error) == n_iter_max:
            axes[0].scatter([init], [error[-1]], color=color, alpha=alpha, marker="x")
        else:
            axes[0].scatter([init], [error[-1]], color=color, alpha=alpha, marker="o")

        axes[1].semilogy(error, color=color, alpha=alpha, zorder=zorder)
        ymax = max(error[1], ymax)

    axes[0].set_xlabel("Initialisation")
    axes[0].set_ylabel("Error")
    axes[1].set_ylim(top=ymax)
    axes[1].set_ylim(top=ymax)

    axes[1].set_xlabel("Iteration")
    axes[1].set_ylabel("Error (Log scale)")

    custom_lines = [
        Line2D([0], [0], marker="o", alpha=1, color="k", linewidth=0),
        Line2D([0], [0], marker="x", alpha=1, color="k", linewidth=0),
        Line2D([0], [0], marker="s", alpha=1, color=plt.rcParams["axes.prop_cycle"].by_key()["color"][1], linewidth=0,),
        Line2D(
            [0], [0], marker="s", alpha=0.5, color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0], linewidth=0,
        ),
    ]

    fig.legend(
        custom_lines,
        ["Converged", "Did not converge", "Lowest final error", "Other runs"],
        ncol=2,
        bbox_to_anchor=(0.5, -0.1),
        loc="lower center",
    )
    return fig, axes


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
def percentage_variation_plot(
    cp_tensor, dataset=None, method="model", ax=None,
):
    """Bar chart showing the percentage of variation explained by each of the components.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray or xarray.DataArray
        Dataset to compare with, only needed if ``method="data"`` or ``method="both"``.
    model : {"model", "data", "both"} (default="model")
        Whether the percentage variation should be computed based on the model, data or both.
    ax : matplotlib axes
        Axes to draw the plot in

    Returns
    -------
    matplotlib axes
        Axes with the plot in

    Examples
    --------

    By default, we get the percentage of variation in the model each component explains

    .. plot::
        :context: close-figs
        :include-source:

        >>> from tlviz.visualisation import percentage_variation_plot
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor, dataset = simulated_random_cp_tensor(shape=(5,10,15), rank=3, noise_level=0.5, seed=0)
        >>> percentage_variation_plot(cp_tensor)
        <AxesSubplot:xlabel='Component number', ylabel='Percentage variation explained [%]'>
        >>> plt.show()

    We can also get the percentage of variation in the data that each component explains

    .. plot::
        :context: close-figs
        :include-source:

        >>> from tlviz.visualisation import percentage_variation_plot
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor, dataset = simulated_random_cp_tensor(shape=(5,10,15), rank=3, noise_level=0.5, seed=0)
        >>> percentage_variation_plot(cp_tensor, dataset, method="data")
        <AxesSubplot:xlabel='Component number', ylabel='Percentage variation explained [%]'>
        >>> plt.show()

    Or both the variation in the data and in the model

    .. plot::
        :context: close-figs
        :include-source:

        >>> from tlviz.visualisation import percentage_variation_plot
        >>> from tlviz.data import simulated_random_cp_tensor
        >>> import matplotlib.pyplot as plt
        >>> cp_tensor, dataset = simulated_random_cp_tensor(shape=(5,10,15), rank=3, noise_level=0.5, seed=0)
        >>> percentage_variation_plot(cp_tensor, dataset, method="both")
        <AxesSubplot:xlabel='Component number', ylabel='Percentage variation explained [%]'>
        >>> plt.show()
    """
    if ax is None:
        ax = plt.gca()

    labels = {"data": "Percentage of data", "model": "Percentage of model"}
    variation = factor_tools.percentage_variation(cp_tensor, dataset, method=method)
    if method == "both":
        data_var, model_var = variation
        ax.bar(np.arange(len(data_var)) - 0.2, data_var, width=0.4, label=labels["data"])
        ax.bar(np.arange(len(model_var)) + 0.2, model_var, width=0.4, label=labels["model"])

    else:
        ax.bar(range(len(variation)), variation, label=labels[method])

    ax.legend()
    ax.set_xlabel("Component number")
    ax.set_ylabel("Percentage variation explained [%]")
    return ax



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

import numpy as np
import pandas as pd
import scipy.stats as stats

from ._module_utils import is_dataframe, is_iterable, is_xarray
from ._tl_utils import _handle_tensorly_backends_cp, _handle_tensorly_backends_dataset, _SINGLETON
from ._xarray_wrapper import is_labelled_dataset
from .utils import _alias_mode_axis, cp_to_tensor

_LEVERAGE_NAME = "Leverage score"
_SLABWISE_SSE_NAME = "Slabwise SSE"

__all__ = [
    "compute_slabwise_sse",
    "compute_leverage",
    "compute_outlier_info",
    "get_leverage_outlier_threshold",
    "get_slabwise_sse_outlier_threshold",
]


def _compute_leverage(factor_matrix):
    A = factor_matrix
    leverage = A @ np.linalg.solve(A.T @ A, A.T)
    return np.diag(leverage)


def _compute_slabwise_sse(estimated, true, normalise=True, axis=0):
    if not is_iterable(axis):
        axis = {axis}
    axis = set(axis)

    reduction_axis = tuple(i for i in range(true.ndim) if i not in axis)
    SSE = ((estimated - true) ** 2).sum(axis=reduction_axis)
    if normalise:
        return SSE / SSE.sum()
    else:
        return SSE


@_alias_mode_axis()
@_handle_tensorly_backends_dataset("true", _SINGLETON)
@_handle_tensorly_backends_dataset("estimated", None)
def compute_slabwise_sse(estimated, true, normalise=True, mode=0, axis=None):
    r"""Compute the (normalised) slabwise SSE along the given mode(s).

    For a tensor, :math:`\mathcal{X}`, and an estimated tensor :math:`\hat{\mathcal{X}}`,
    we compute the :math:`i`-th normalised slabwise residual as

    .. math::
        r_i = \frac{\sum_{jk} \left(x_{ijk} - \hat{x}_{ijk}\right)^2}
                   {\sum_{ijk} \left(x_{ijk} - \hat{x}_{ijk}\right)^2}.

    The residuals can measure how well our decomposition fits the different
    sample. If a sample, :math:`i`, has a high residual, then that indicates that
    the model is not able to describe its behaviour.

    Parameters
    ----------
    estimated : xarray or numpy array
        Estimated dataset, if this is an xarray, then the output is too.
    true : xarray or numpy array
        True dataset, if this is an xarray, then the output is too.
    normalise : bool
        Whether the SSE should be scaled so the vector sums to one.
    mode : int or iterable of ints
        Mode (or modes) that the SSE is computed across (i.e. these are not the ones summed over).
        The output will still have these axes.
    axis : int or iterable of ints (optional)
        Alias for mode. If this is set, then no value for mode can be given

    Returns
    -------
    slab_sse : xarray or numpy array
        The (normalised) slabwise SSE, if true tensor input is an xarray array,
        then the returned tensor is too.
    """
    # TODOC: example for compute_slabwise_sse
    # Check that dimensions match up.
    if is_xarray(estimated) and is_xarray(true):
        if estimated.dims != true.dims:
            raise ValueError(
                f"Dimensions of estimated and true tensor must be equal,"
                f" they are {estimated.dims} and {true.dims}, respectively."
            )
        for dim in estimated.dims:
            if len(true.coords[dim]) != len(estimated.coords[dim]):
                raise ValueError(
                    f"The dimension {dim} has different length for the true and estiamted tensor. "
                    f"The true tensor has length {len(true.coords[dim])} and the estimated tensor "
                    f"has length {len(estimated.coords[dim])}."
                )
            if not all(true.coords[dim] == estimated.coords[dim]):
                raise ValueError(f"The dimension {dim} has different coordinates for the true and estimated tensor.")
    elif is_dataframe(estimated) and is_dataframe(true):
        if estimated.columns != true.columns:
            raise ValueError("Columns of true and estimated matrix must be equal")
        if estimated.index != true.index:
            raise ValueError("Index of true and estimated matrix must be equal")

    slab_sse = _compute_slabwise_sse(estimated, true, normalise=normalise, axis=mode)
    if is_labelled_dataset(true):
        slab_sse.name = _SLABWISE_SSE_NAME
    return slab_sse


@_handle_tensorly_backends_dataset("factor_matrix", _SINGLETON)
def compute_leverage(factor_matrix):
    r"""Compute the leverage score of the given factor matrix.

    The leverage score is a measure of how much "influence" a slab (often representing a sample)
    has on a tensor factorisation model. To compute the leverage score for the different slabs,
    we only need the factor matrix for the selected mode. If the selected mode is represented
    by :math:`\mathbf{A}`, then the leverage score is defined as

    .. math::

        h_i = \left[\mathbf{A} \left(\mathbf{A}^T \mathbf{A}\right)^{-1} \mathbf{A}^T\right]_{ii},

    that is, the :math:`i`-th diagonal entry of the matrix
    :math:`\mathbf{A} \left(\mathbf{A}^T \mathbf{A}\right)^{-1} \mathbf{A}^T`.
    If a given slab, :math:`i`, has a high leverage score, then it likely has a strong
    influence on the model. A good overview of the leverage score is :cite:p:`velleman1981efficient`.

    The leverage scores sums to the number of components for our model and is always between 0 and 1.
    Moreover, if a data point has a leverage score equal to 1, then one component is solely "devoted"
    to modelling that data point, and removing the corresponding row from :math:`A` will reduce the
    rank of :math:`A` by 1 :cite:p:`belsley1980regression`.

    A way of interpreting the leverage score is as a measure of how "similar" a data point
    is to the rest. If a row of :math:`A` is equal to the average row of :math:`A`, then its leverage
    score would be equal to :math:`\frac{1}{I}`. Likewise, if a data point has a leverage of 1, then
    no other data points have a similar model representation. If a data point has a leverage of 0.5,
    then there is one other data point (in some weighted sense) with a similar model representation,
    and a leverage of 0.2 means that there are five other data points with a similar model representation
    :cite:p:`huber2009robust`.

    If the factor matrix is a dataframe, then the output is also a dataframe with that index. Otherwise,
    the output is a NumPy array.

    Parameters
    ----------
    factor_matrix : DataFrame or numpy array
        The factor matrix whose leverage we compute

    Returns
    -------
    leverage : DataFrame or numpy array
        The leverage scores, if the input is a dataframe, then the index is preserved.

    .. note::

        The leverage score is related to the Hotelling T2-statistic (or D-statistic), which
        is equal to a scaled version of leverage computed based on centered factor matrices.

    Examples
    --------

    In this example, we compute the leverage of a random factor matrix

    >>> import numpy as np
    >>> from tlviz.outliers import compute_leverage
    >>> rng = np.random.default_rng(0)
    >>> A = rng.standard_normal(size=(5, 2))
    >>> leverage_scores = compute_leverage(A)
    >>> for index, leverage in enumerate(leverage_scores):
    ...     print(f"Sample {index} has leverage score {leverage:.2f}")
    Sample 0 has leverage score 0.04
    Sample 1 has leverage score 0.23
    Sample 2 has leverage score 0.50
    Sample 3 has leverage score 0.59
    Sample 4 has leverage score 0.64
    """
    leverage = _compute_leverage(factor_matrix)

    if is_dataframe(factor_matrix):
        return pd.DataFrame(leverage.reshape(-1, 1), columns=[_LEVERAGE_NAME], index=factor_matrix.index)
    else:
        return leverage


@_alias_mode_axis()
@_handle_tensorly_backends_dataset("true_tensor", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
def compute_outlier_info(cp_tensor, true_tensor, normalise_sse=True, mode=0, axis=None):
    """Compute the leverage score and (normalised) slabwise SSE along one axis.

    These metrics are often plotted against each other to discover outliers.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    true_tensor : xarray or numpy array
        Dataset that cp_tensor is fitted against.
    normalise_sse : bool
        If true, the slabwise SSE is scaled so it sums to one.
    mode : int
        The mode to compute the outlier info across.
    axis : int (optional)
        Alias for mode. If this is set, then no value for mode can be given.

    Returns
    -------
    DataFrame
        Dataframe with two columns, "Leverage score" and "Slabwise SSE".

    See Also
    --------
    compute_leverage : More information about the leverage score is given in this docstring
    compute_slabwise_sse : More information about the slabwise SSE is given in this docstring
    get_leverage_outlier_threshold : Cutoff for selecting potential outliers based on the leverage
    compute_slabwise_sse : Cutoff for selecting potential outliers based on the slabwise SSE
    """
    # TODOC: Example for compute_outlier_info
    # Add whether suspicious based on rule-of-thumb cutoffs as boolean columns
    leverage = compute_leverage(cp_tensor[1][mode])

    estimated_tensor = cp_to_tensor(cp_tensor)
    slab_sse = compute_slabwise_sse(estimated_tensor, true_tensor, normalise=normalise_sse, mode=mode)
    if is_xarray(slab_sse):
        slab_sse = pd.DataFrame(slab_sse.to_series())

    leverage_is_labelled = is_dataframe(leverage)
    sse_is_labelled = is_dataframe(slab_sse)
    if (leverage_is_labelled and not sse_is_labelled) or (not leverage_is_labelled and sse_is_labelled):
        raise ValueError(
            "If `cp_tensor` is labelled (factor matrices are dataframes), then"
            "`true_tensor` should be an xarray object and vice versa."
        )
    elif not leverage_is_labelled and not sse_is_labelled:
        return pd.DataFrame({_LEVERAGE_NAME: leverage, _SLABWISE_SSE_NAME: slab_sse})
    elif leverage_is_labelled and not all(slab_sse.index == leverage.index):
        raise ValueError("The indices of the labelled factor matrices does not match up with the xarray dataset")

    results = pd.concat([leverage, slab_sse], axis=1)
    results.columns = [_LEVERAGE_NAME, _SLABWISE_SSE_NAME]
    return results


@_handle_tensorly_backends_dataset("leverage_scores", None)
def get_leverage_outlier_threshold(leverage_scores, method="p_value", p_value=0.05):
    """Compute threshold for detecting possible outliers based on leverage.

    **Huber's heuristic for selecting outliers**

    In Robust Statistics, Huber :cite:p:`huber2009robust` shows that that if the leverage score,
    :math:`h_i`, of a sample is equal to :math:`1/r` and we duplicate that sample, then its leverage
    score will be equal to :math:`1/(1+r)`. We can therefore, think of of the reciprocal of the
    leverage score, :math:`1/h_i`, as the number of similar samples in the dataset. Following this
    logic, Huber recommends two thresholds for selecting outliers: 0.2 (which we name ``"huber low"``)
    and 0.5 (which we name ``"huber high"``).

    **Hoaglin and Welch's heuristic for selecting outliers**

    In :cite:p:`hoaglin1978hat`, :cite:authors:`hoaglin1978hat` state that :math:`2r/n` is a good cutoff
    for selecting samples that may be outliers. This choice is elaborated in :cite:p:`belsley1980regression`
    (page 17), where :cite:authors:`belsley1980regression` also propose :math:`3r/n` as a cutoff when
    :math:`r < 6` and :math:`n-r > 12`. They also defend thee cut-offs by proving that if the factor matrices
    are normally distributed, then :math:`(n - r)[h_i - (1/n)]/[(1 - h_i)(r - 1)]` follows a Fisher
    distribution with :math:`(r-1)` and :math:`(n-r)` degrees of freedom. While the factor matrix
    seldomly follows a normal distribution, :cite:authors:`belsley1980regression` still argues that this
    can be a good starting point for cut-off values of suspicious data points. Based on reasonable choices for
    :math:`n` and :math:`r`, they arive at the heuristics above.

    **Leverage p-value**

    Another way to select ouliers is also based on the findings by :cite:authors:`belsley1980regression`.
    We can use the transformation into a Fisher distributed variable (assuming that the factor elements
    are drawn from a normal distribution), to compute cut-off values based on a p-value. The elements of
    the factor matrices are seldomly normally distributed, so this is also just a rule-of-thumb.

    .. note::

        Note also that we, with bootstrap estimation, have found that this p-value is only valid for
        large number of components. For smaller number of components, the false positive rate will be higher
        than the specified p-value, even if the components follow a standard normal distribution (see example below).

    **Hotelling's T2 statistic**

    Yet another way to estimate a p-value is via Hotelling's T-squared statistic :cite:p:`jackson1980principal`
    (see also :cite:p:`nomikos1995multivariate`). The key here is to notice that if the factor matrices are
    normally distributed with zero mean, then the leverage is equivalent to a scaled version of the Hotelling's
    T-squared statistic. This is commonly used in PCA, where the data often is centered beforehand, which leads
    to components with zero mean (in the mode the data is centered across). Again, note that the elements of the
    factor matrices are seldomly normally distributed, so this is also just a rule-of-thumb.

    .. note::

        Note also that we, with bootstrap estimation, have found that this p-value is not valid for
        large numbers of components. In that case, the false positive rate will be higher than the specified
        p-value, even if the components follow a standard normal distribution (see example below).

    Parameters
    ----------
    leverage_scores : np.ndarray or pd.DataFrame
    method : {"huber lower", "huber higher", "hw lower", "hw higher", "p-value", "hotelling"}
    p_value : float (optional, default=0.05)
        If ``method="p-value"``, then this is the p-value used for the cut-off.

    Returns
    -------
    threshold : float
        Threshold value, data points with a leverage score larger than the threshold are suspicious
        and may be outliers.

    Examples
    --------

    **The leverage p-value is only accurate with many components:**
    Here, we use Monte-Carlo estimation to demonstrate that the p-value derived in :cite:p:`belsley1980regression`
    is valid only for large number of components.

    We start by importing some utilities

    >>> import numpy as np
    >>> from scipy.stats import bootstrap
    >>> from tlviz.outliers import compute_leverage, get_leverage_outlier_threshold

    Here, we create a function that computes the false positive rate

    >>> def compute_false_positive_rate(n, d, p_value):
    ...     X = np.random.standard_normal((n, d))
    ...
    ...     h = compute_leverage(X)
    ...     th = get_leverage_outlier_threshold(h, method="p-value", p_value=p_value)
    ...     return (h > th).mean()

    >>> np.random.seed(0)
    >>> n_samples = 1_000
    >>> leverages = [compute_false_positive_rate(10, 2, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(leverages, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0815, 0.0897]

    We see that the false positive rate is almost twice what we prescribe (0.05). However, if we increase
    the number of components, then the false positive rate improves

    >>> leverages = [compute_false_positive_rate(10, 9, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(leverages, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0468, 0.0554]

    This indicates that the false positive rate is most accurate when the number of components is equal
    to the number of samples - 1. We can increase the number of samples to assess this conjecture

    >>> leverages = [compute_false_positive_rate(100, 9, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(leverages, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0558, 0.0581]

    The increase in the false positive rate supports the conjecture that :cite:author:`belsley1980regression`'s
    method for computing the p-value is accurate only when the number of components is high. Still, it is
    important to remember that the original assumptions (normally distributed components) is seldomly satisfied
    also, so this method is only a rule-of-thumb and can still be useful.

    **Hotelling's T-squared statistic requires few components or many samples:**
    Here, we use Monte-Carlo estimation to demonstrate that the Hotelling T-squared statistic is only valid with
    many samples.

    >>> def compute_hotelling_false_positive_rate(n, d, p_value):
    ...     X = np.random.standard_normal((n, d))
    ...
    ...     h = compute_leverage(X)
    ...     th = get_leverage_outlier_threshold(h, method="hotelling", p_value=p_value)
    ...     return (h > th).mean()

    We set the simulation parameters and the seed

    >>> np.random.seed(0)
    >>> n_samples = 1_000
    >>> fprs = [compute_hotelling_false_positive_rate(10, 2, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(fprs, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0492, 0.0568]

    However, if we increase the number of components, then the false positive rate becomes to large

    >>> fprs = [compute_hotelling_false_positive_rate(10, 5, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(fprs, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0738, 0.0833]

    But if we increase the number of samples, then the estimate is good again

    >>> fprs = [compute_hotelling_false_positive_rate(100, 5, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(fprs, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0494, 0.0515]
    """
    num_samples = int(round(len(leverage_scores)))
    num_components = int(round(np.sum(leverage_scores)))

    method = method.lower()
    if method == "huber lower":
        return 0.2
    elif method == "huber higher":
        return 0.5
    elif method == "hw lower":
        return 2 * num_components / num_samples
    elif method == "hw higher":
        return 3 * num_components / num_samples
    elif method == "p-value":
        n, p = num_samples, num_components

        if p <= 1:
            raise ValueError("Cannot use P-value when there is only one component.")
        if n <= p:
            raise ValueError("Cannot use P-value when there are fewer samples than components.")

        # TODO: Try with n - p - 1, and try maths too compare with hotelling
        F = stats.f.isf(p_value, p - 1, n - p)
        F_scale = F * (p - 1) / (n - p)
        # Solve the equation (h + (1/n)) / (1 - h) = F_scale:
        return (F_scale + (1 / n)) / (1 + F_scale)
    elif method == "hotelling":
        I, R = num_samples, num_components
        if I <= R + 1:
            raise ValueError("Cannot use Hotelling P-value when there are fewer samples than components minus one.")

        F = stats.f.isf(p_value, R, I - R - 1)
        B = F * (R / (I - R - 1)) / (1 + (R / (I - R - 1)) * F)
        # Remove the square compared to Nomikos & MacGregor since the leverage is:
        #  A(AtA)^-1 At,
        # not
        #  A(AtA / (I-1))^-1 At
        return B * (I - 1) / I
    else:
        raise ValueError(
            "Method must be one of 'huber lower', 'huber higher', 'hw lower' or 'hw higher', "
            + f"'p-value', or 'hotelling' not {method}"
        )


@_handle_tensorly_backends_dataset("slab_sse", None)
def get_slabwise_sse_outlier_threshold(slab_sse, method="p-value", p_value=0.05, ddof=1):
    r"""Compute rule-of-thumb threshold values for suspicious residuals.

    One way to determine possible outliers is to examine how well the model describes
    the different data points. A standard way of measuring this, is by the slab-wise
    sum of squared errors (slabwise SSE), which is the sum of squared error for each
    data point.

    There is, unfortunately, no guaranteed way to detect outliers automatically based
    on the residuals. However, if the noise is normally distributed, then the residuals
    follow a scaled chi-squared distribution. Specifically, we have that
    :math:`\text{SSE}_i^2 \sim g\chi^2_h`, where :math:`g = \frac{\sigma^2}{2\mu}`,
    :math:`h = \frac{\mu}{g} = \frac{2\mu^2}{\sigma^2}`, and :math:`\mu` is the
    average slabwise SSE and :math:`\sigma^2` is the variance of the slabwise
    SSE :cite:p:`box1954some`.

    Another rule-of-thumb follows from :cite:p:`naes2002user` (p. 187), which states
    that two times the standard deviation of the slabwise SSE can be used for
    determining data points with a suspiciously high residual.

    Parameters
    ----------
    slab_sse : np.ndarray or pd.DataFrame
    method : {"two_sigma", "p-value"}
    p_value : float (optional, default=0.05)
        If ``method="p-value"``, then this is the p-value used for the cut-off.

    Returns
    -------
    threshold : float
        Threshold value, data points with a higher SSE than the threshold are suspicious
        and may be outliers.

    Examples
    --------
    Here, we see that the p-value gives a good cutoff if the noise is normally distributed

    We start by importing the tools we'll need

    >>> import numpy as np
    >>> from scipy.stats import bootstrap
    >>> from tlviz.outliers import compute_slabwise_sse, get_slabwise_sse_outlier_threshold
    >>> from tlviz.utils import cp_to_tensor

    Then, we create a function to compute the false positive rate. This will be useful for our
    bootstrap estimate for the true false positive rate.

    >>> def compute_false_positive_rate(shape, num_components, p_value):
    ...     A = np.random.standard_normal((shape[0], num_components))
    ...     B = np.random.standard_normal((shape[1], num_components))
    ...     C = np.random.standard_normal((shape[2], num_components))
    ...
    ...     X = cp_to_tensor((None, [A, B, C]))
    ...     noisy_X = X + np.random.standard_normal(shape)*5
    ...
    ...
    ...
    ...     sse = compute_slabwise_sse(X, noisy_X)
    ...     th = get_slabwise_sse_outlier_threshold(sse, method="p-value", p_value=p_value)
    ...     return (sse > th).mean()

    Finally, we estimate the 95% confidence interval of the false positive rate to validate
    that it is approximately correct.

    >>> np.random.seed(0)
    >>> n_samples = 1_000
    >>> slab_sse = [compute_false_positive_rate((20, 20, 10), 5, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(slab_sse, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0434, 0.0474]

    We see that the 95% confidence interval lies just below our goal of 0.05! Let's also try
    with a false positive rate of 0.1

    >>> slab_sse = [compute_false_positive_rate((20, 20, 10), 5, 0.1) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(slab_sse, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0972, 0.1021]

    Here we see that the false positive rate is sufficiently estimated. It may have been too low
    above since we either did not have enough samples in the first mode (which we compute) the
    false positive rate for). With only 20 samples, it will be difficult to correctly estimate a
    false positive rate of 0.05. If we increase the number of samples to 200 instead, we see that
    the false positive rate is within our expected bounds.

    >>> slab_sse = [compute_false_positive_rate((200, 20, 10), 5, 0.05) for _ in range(n_samples)],
    >>> fpr_low, fpr_high = bootstrap(slab_sse, np.mean).confidence_interval
    >>> print(f"95% confidence interval for the false positive rate: [{fpr_low:.4f}, {fpr_high:.4f}]")
    95% confidence interval for the false positive rate: [0.0494, 0.0507]
    """
    std = np.std(slab_sse, ddof=ddof)
    mean = np.mean(slab_sse)
    if method == "two sigma":
        return std * 2
    elif method == "p-value":
        g = std * std / (2 * mean)
        h = mean / g
        return stats.chi2.isf(p_value, h) * g
    else:
        raise ValueError(f"Method must be one of 'two sigma' and 'p-value', not '{method}'.")



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

import zipfile
from io import BytesIO
from pathlib import Path

import numpy as np
import pandas as pd
import requests
import xarray as xr
from scipy.io import loadmat

from .utils import cp_to_tensor

__all__ = ["load_aminoacids", "load_oslo_city_bike", "download_city_bike"]

DATASET_PARENT = Path(__file__).parent / "datasets"
DOWNLOADED_PARENT = DATASET_PARENT / "downloads"


class RemoteZip:
    def __init__(self, url):
        req = requests.get(url)
        self.zip = zipfile.ZipFile(BytesIO(req.content))

    @property
    def contents(self):
        return [f.filename for f in self.zip.infolist()]

    def extract_file(self, filename):
        file_contents = self.zip.read(filename)
        data = BytesIO(file_contents)
        return data

    def extract_all(self):
        return {filename: self.extract_file(filename) for filename in self.contents}


def load_aminoacids(save_data=True):
    """Load the Aminoacids dataset from "PARAFAC. Tutorial and applications" by Rasmus Bro :cite:p:`bro1997parafac`.

    This is a fluoresence spectroscopy dataset well suited for PARAFAC analysis.
    The data stems from five samples with different amounts of the aminoacids tyrosine, tryptophan and phenylalanine.

    See here for more information about the data:
        http://models.life.ku.dk/Amino_Acid_fluo
    Or archived version here:
        https://web.archive.org/web/20210413050155/http://models.life.ku.dk/Amino_Acid_fluo

    Parameters
    ----------
    save_data : bool (default=True)
        If ``True``, then the dataset is saved in ``tlviz.data.DOWNLOAD_PARENT`` to avoid having to download
        it in the future.

    Returns
    -------
    xarray.DataArray
        The Aminoacids dataset
    """
    print("Loading Aminoacids dataset from:")
    print(
        "Bro, R, PARAFAC: Tutorial and applications, Chemometrics and Intelligent Laboratory Systems,"
        + " 1997, 38, 149-171"
    )

    filename = DOWNLOADED_PARENT / "aminoacids.nc4"
    if filename.is_file():
        return xr.open_dataarray(filename)

    aminoacids_zip = RemoteZip("http://models.life.ku.dk/sites/default/files/Amino_Acid_fluo.zip")
    matlab_variables = loadmat(aminoacids_zip.extract_file("amino.mat"))

    I, K, J = matlab_variables["DimX"].squeeze().astype(int)
    X = matlab_variables["X"].reshape(I, J, K)
    emission_frequencies = matlab_variables["EmAx"].squeeze()
    excitation_frequencies = matlab_variables["ExAx"].squeeze()
    coords_dict = {
        "Sample": list(range(I)),
        "Emission frequency": emission_frequencies,
        "Excitation frequency": excitation_frequencies,
    }
    dims = ["Sample", "Excitation frequency", "Emission frequency"]

    data = xr.DataArray(X, dims=dims, coords=coords_dict)

    if save_data:
        DOWNLOADED_PARENT.mkdir(exist_ok=True, parents=True)
        data.to_netcdf(filename)

    return data


def load_oslo_city_bike():
    """Download bike data from the bike sharing system in Oslo 2020-2021.

    The dataset is a five-way tensor with modes:

     * Bike station id
     * Year
     * Month
     * Day of week
     * Hour of day

    where the elements corresponds to the number of trips ending in a given
    time period and station. For example, the index ``377, 2020, 5, 1, 13`` corresponds
    to the number of trips that ended at station 377 at 1pm (CET time) on Mondays in May of 2020.

    The data was collected using the open API of https://oslobysykkel.no/en/open-data.

    Returns
    -------
    xarray.DataArray
        Data array whose data is the dataset and coordinates are bike station ID, year, month, day of week
        and hour of day. There are also three metadata-coordinates along the "Bike station ID"-axis containing
        latitudes, longitudes and station names.
    """
    nc4_file = DATASET_PARENT / "oslo_bike.nc4"
    return xr.load_dataarray(nc4_file)


def download_city_bike(source="oslobysykkel.no", years=(2020, 2021)):
    """Download bike data from the bike sharing system in Oslo.

    The dataset is a five-way tensor with modes:

     * Bike station id
     * Year
     * Month
     * Day of week
     * Hour of day

    where the elements corresponds to the number of trips ending in a given
    time period and station. For example, the index ``377, 2020, 5, 1, 13`` corresponds
    to the number of trips that ended at station 377 at 1pm (CET time) on Mondays in May of 2020.

    The data was collected using the open urbansharing API.

    Parameters
    ----------
    source : str
        String to download data from. The API endpoint becomes:
        https://data.urbansharing.com/{source}/trips/v1/{year}/{month:02d}.csv
        Some end-points that are known to work are:

         * oslobysykkel.no
         * bergenbysykkel.no
         * trondheimbysykkel.no

    years : iterable of int
        The years to download data from.

    Returns
    -------
    xarray.DataArray
        Data array whose data is the dataset and coordinates are bike station ID, year, month, day of week
        and hour of day. There are also three metadata-coordinates along the "Bike station ID"-axis containing
        latitudes, longitudes and station names.
    """
    end_data = []

    lat = {}
    lon = {}
    name = {}
    for year in years:
        print(f"Loading {year}", flush=True)
        for month in range(12):
            month += 1
            df = pd.read_csv(f"https://data.urbansharing.com/{source}/trips/v1/{year}/{month:02d}.csv")
            # station_names = sorted(set(df["start_station_name"]) | set(df["end_station_name"]))

            df["ended_at"] = df["ended_at"].map(lambda x: x if "." in x else ".000000+".join(x.split("+")))
            df["ended_at"] = pd.to_datetime(df["ended_at"], format="%Y-%m-%d %H:%M:%S.%f+00:00")
            df["ended_at"] = df["ended_at"].dt.tz_localize("UTC")
            df["ended_at"] = df["ended_at"].dt.tz_convert("CET")

            end_time = df["ended_at"].dt

            df["trip"] = 1
            df["Day"] = end_time.day
            df["Day of week"] = end_time.dayofweek
            df["Hour"] = end_time.hour
            df["Month"] = end_time.month
            df["Year"] = end_time.year
            df = df.rename({"end_station_id": "End station ID"}, axis=1)

            for _, row in df.iterrows():
                lat[row["End station ID"]] = row["end_station_latitude"]
                lon[row["End station ID"]] = row["end_station_longitude"]
                name[row["End station ID"]] = row["end_station_name"]

            end = df.groupby(["End station ID", "Year", "Month", "Day of week", "Hour"]).sum()[["trip"]]
            end_data.append(end)

    grouped = pd.concat(end_data).groupby(level=(0, 1, 2, 3, 4)).sum()

    dataset = xr.Dataset.from_dataframe(grouped).to_array().squeeze()

    # Drop trips that started in the last year of interest and ended the year after
    unwanted_years = set(dataset.coords["Year"].values) - set(years)
    for year in unwanted_years:
        dataset = dataset.drop(year, "Year")

    # Set equal to 0 for all hours with no trips
    dataset.values[np.isnan(dataset.values)] = 0

    # Add station metadata
    lat = [lat[station_id.item()] for station_id in dataset.coords["End station ID"]]
    lon = [lon[station_id.item()] for station_id in dataset.coords["End station ID"]]
    name = [name[station_id.item()] for station_id in dataset.coords["End station ID"]]
    dataset = xr.DataArray(
        dataset.data,
        coords={
            **{key: value for key, value in dataset.coords.items() if key != "variable"},
            "lat": (("End station ID",), lat),
            "lon": (("End station ID",), lon),
            "name": (("End station ID",), name),
        },
        dims=dataset.dims,
        name="Bike trips",
    )
    return dataset


def simulated_random_cp_tensor(shape, rank, noise_level=0.1, labelled=False, seed=None):
    """Create a random noisy CP tensor.

    The factor matrices and weights have uniformly distributed elements and
    the noise is normally distributed with magnitude ``noise_level * tl.norm(X)``,
    where ``X`` is the datatensor represented by the decomposition.

    Parameters
    ----------
    shape : iterable of ints
        The shape of the data tensor
    rank : int
        The number of components
    noise_level : float
        Relative magnitude of the noise compared to the magnitude of the data.
    seed : {None, int, array_like[ints], np.random.SeedSequence, np.random.BitGenerator, np.random.Generator}
        Seed for numpy random number generator

    Returns
    -------
    cp_tensor
        CP tensor that represents the simulated dataset

    np.ndarray
        Dense tensor with noise added
    """
    rng = np.random.default_rng(seed)
    weights = rng.random(size=rank)
    factors = [rng.random(size=(length, rank)) for length in shape]
    if labelled:
        for i, factor in enumerate(factors):
            factors[i] = pd.DataFrame(factor)
            factors[i].index.name = f"Mode {i}"
    cp = weights, factors

    X = cp_to_tensor(cp)
    noise = rng.standard_normal(size=shape)
    X_noisy = X + np.linalg.norm(X) * noise_level * noise / np.linalg.norm(noise)

    return cp, X_noisy


# TODO: Add more example datasets
# TODO: Enron data
# TOTEST: data.py
